# -*- coding: utf-8 -*-
#
#  Qt 4 based frontend to software-properties
#
#  Copyright © 2009 Harald Sitter <apachelogger@ubuntu.com>
#  Copyright © 2007 Canonical Ltd.
#
#  Author: Jonathan Riddell <jriddell@ubuntu.com>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

# FIXME: why does this pass kapps around?????

from __future__ import absolute_import, print_function

import apt_pkg
import tempfile
from gettext import gettext as _
import os

import gi
gi.require_version('PackageKitGlib', '1.0')
from gi.repository import PackageKitGlib as packagekit
from gi.repository import Gio

from PyQt5 import uic
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

import softwareproperties
import softwareproperties.distro
from softwareproperties.SoftwareProperties import SoftwareProperties
import softwareproperties.SoftwareProperties
from .I18nHelper import *

from aptsources.sourceslist import SourceEntry
from .DialogEdit import DialogEdit
from .DialogAdd import DialogAdd
from .DialogMirror import DialogMirror
from .CdromProgress import CdromProgress

from UbuntuDrivers import detect
import sys
from functools import partial
import aptsources.distro
import logging

class DetectDriverThread(QThread):
  #WARNING class to run detect_drivers() in separate thread
  #so GUI won't freeze.
  def __init__(self, swprop):
    QThread.__init__(self)
    self.swprop = swprop

  def __del__(self):
    self.wait()

  def run(self):
    self.swprop.detect_drivers()

class SoftwarePropertiesQtUI(QWidget):
  def __init__(self, datadir):
      QWidget.__init__(self)
      uic.loadUi("%s/designer/main.ui" % datadir, self)

class SoftwarePropertiesQt(SoftwareProperties):
  def __init__(self, datadir=None, options=None, parent=None, file=None, attachWinID=None):
    """ Provide a Qt-based graphical user interface to configure
        the used software repositories, corresponding authentication keys
        and update automation """
    SoftwareProperties.__init__(self, options=options, datadir=datadir)

    self.options = options
    self.datadir = datadir

    global kapp
    kapp = QApplication.instance()
    kapp.setWindowIcon(QIcon.fromTheme("applications-other"))

    self.userinterface = SoftwarePropertiesQtUI(datadir)
    self.userinterface.setWindowIcon(QIcon.fromTheme("applications-other"))
    self.userinterface.button_auth_restore.setIcon(QIcon.fromTheme("edit-undo"))
    self.userinterface.button_add_auth.setIcon(QIcon.fromTheme("list-add"))
    self.userinterface.button_auth_remove.setIcon(QIcon.fromTheme("list-remove"))
    self.userinterface.button_remove.setIcon(QIcon.fromTheme("list-remove"))
    self.userinterface.button_edit.setIcon(QIcon.fromTheme("document-edit"))
    self.userinterface.button_add.setIcon(QIcon.fromTheme("list-add"))
    self.userinterface.button_add_cdrom.setIcon(QIcon.fromTheme("media-optical"))
    translate_widget(self.userinterface)
    self.userinterface.show()
    # FIXME: winid not handled
    #if attachWinID is not None:
        #KWindowSystem.setMainWindow(self.userinterface, int(attachWinID))

    # rejected() signal from Close button
    self.userinterface.buttonBox.rejected.connect(self.on_close_button)

    self.userinterface.buttonBox.button(QDialogButtonBox.Reset).setEnabled(False)

    # Put some life into the user interface:
    self.init_server_chooser()
    self.init_popcon()
    self.init_auto_update()
    self.init_release_upgrades()
    self.show_auto_update_level()
    # Setup the key list
    self.init_keys()
    self.show_keys()
    # Setup the ISV sources list
    self.init_isv_sources()
    self.show_isv_sources()
    self.show_cdrom_sources()
    # Setup and show the Additonal Drivers tab
    self.init_drivers()

    # Connect to switch-page before setting initial tab. Otherwise the
    # first switch goes unnoticed.
    self.userinterface.tabWidget.currentChanged.connect(self.tab_switched)

    self.userinterface.checkbutton_source_code.clicked.connect(self.on_checkbutton_source_code_toggled)
    self.userinterface.button_auth_restore.clicked.connect(self.on_restore_clicked)
    self.userinterface.button_add_auth.clicked.connect(self.add_key_clicked)
    self.userinterface.button_auth_remove.clicked.connect(self.remove_key_clicked)
    self.userinterface.checkbutton_popcon.toggled.connect(self.on_checkbutton_popcon_toggled)
    self.userinterface.checkbutton_auto_update.toggled.connect(self.on_auto_update_toggled)
    self.userinterface.combobox_update_interval.currentIndexChanged.connect(self.on_combobox_update_interval_changed)
    self.userinterface.button_remove.clicked.connect(self.on_remove_clicked)
    self.userinterface.button_edit.clicked.connect(self.on_edit_clicked)
    self.userinterface.button_add_cdrom.clicked.connect(self.on_button_add_cdrom_clicked)
    self.userinterface.button_add.clicked.connect(self.on_add_clicked)
    self.userinterface.treeview_sources.itemChanged.connect(self.on_isv_source_toggled)
    self.userinterface.treeview_sources.itemClicked.connect(self.on_treeview_sources_cursor_changed)
    self.userinterface.treeview_cdroms.itemChanged.connect(self.on_cdrom_source_toggled)
    self.userinterface.treeview2.itemClicked.connect(self.on_treeview_keys_cursor_changed)

    self.button_close = self.userinterface.buttonBox.button(QDialogButtonBox.Close)
    self.button_close.setIcon(QIcon.fromTheme("dialog-close"))
    self.button_revert = self.userinterface.buttonBox.button(QDialogButtonBox.Reset)
    self.button_revert.setIcon(QIcon.fromTheme("edit-undo"))
    self.button_revert.clicked.connect(self.on_button_revert_clicked)

    self.init_distro()
    self.show_distro()

    if options and options.open_tab:
      self.userinterface.tabWidget.setCurrentIndex(int(options.open_tab))

  def tab_switched(self):
    # On the additional drivers page, don't show the backend revert button.
    if self.userinterface.tabWidget.currentIndex() == 4: #vbox_drivers is 4
      self.button_revert.setVisible(False)
      if not self.detect_called:
        #WARNING detect_drivers() runs in separate thread
        #in DetectDriverThread class so GUI won't freeze
        #after finish show_drivers() has to run in main thread because it updates the GUI
        self.detect_driver_thread = DetectDriverThread(self)
        self.detect_driver_thread.finished.connect(self.show_drivers)
        self.detect_driver_thread.start()

    else:
      self.button_revert.setVisible(True)

  def init_popcon(self):
    """ If popcon is enabled show the statistics tab and an explanation
        corresponding to the used distro """
    is_helpful = self.get_popcon_participation()
    if is_helpful != None:
      text = softwareproperties.distro.get_popcon_description(self.distro)
      text = text.replace("\n", "<br />") #silly GTK mixes HTML and normal white space
      self.userinterface.label_popcon_desc.setText(text)
      self.userinterface.checkbutton_popcon.setChecked(is_helpful)
    else:
      self.userinterface.tabWidget.removeTab(5)

  def init_server_chooser(self):
    """ Set up the widgets that allow to choose an alternate download site """
    # nothing to do here, set up signal in show_distro()
    pass

  def init_release_upgrades(self):
    " setup the widgets that allow configuring the release upgrades "
    i = self.get_release_upgrades_policy()

    self.userinterface.combobox_release_upgrades.setCurrentIndex(i)
    self.userinterface.combobox_release_upgrades.currentIndexChanged.connect(self.on_combobox_release_upgrades_changed)

  def init_auto_update(self):
    """ Set up the widgets that allow to configure the update automation """
    # this maps the key (combo-box-index) to the auto-update-interval value
    # where (-1) means, no key
    self.combobox_interval_mapping = { 0 : 1,
                                       1 : 2,
                                       2 : 7,
                                       3 : 14 }

    self.userinterface.combobox_update_interval.insertItem(99, _("Daily"))
    self.userinterface.combobox_update_interval.insertItem(99, _("Every two days"))
    self.userinterface.combobox_update_interval.insertItem(99, _("Weekly"))
    self.userinterface.combobox_update_interval.insertItem(99, _("Every two weeks"))

    update_days = self.get_update_interval()

    # If a custom period is defined add a corresponding entry
    if not update_days in self.combobox_interval_mapping.values():
        if update_days > 0:
            self.userinterface.combobox_update_interval.insertItem(99, _("Every %s days") % update_days)
            self.combobox_interval_mapping[4] = update_days

    for key in self.combobox_interval_mapping:
      if self.combobox_interval_mapping[key] == update_days:
        self.userinterface.combobox_update_interval.setCurrentIndex(key)
        break

    if update_days >= 1:
      self.userinterface.checkbutton_auto_update.setChecked(True)
      self.userinterface.combobox_update_interval.setEnabled(True)
      self.userinterface.radiobutton_updates_inst_sec.setEnabled(True)
      self.userinterface.radiobutton_updates_download.setEnabled(True)
      self.userinterface.radiobutton_updates_notify.setEnabled(True)
    else:
      self.userinterface.checkbutton_auto_update.setChecked(False)
      self.userinterface.combobox_update_interval.setEnabled(False)
      self.userinterface.radiobutton_updates_inst_sec.setEnabled(False)
      self.userinterface.radiobutton_updates_download.setEnabled(False)
      self.userinterface.radiobutton_updates_notify.setEnabled(False)

    self.userinterface.radiobutton_updates_download.toggled.connect(self.set_update_automation_level)
    self.userinterface.radiobutton_updates_inst_sec.toggled.connect(self.set_update_automation_level)
    self.userinterface.radiobutton_updates_notify.toggled.connect(self.set_update_automation_level)

  def show_auto_update_level(self):
    """Represent the level of update automation in the user interface"""
    level = self.get_update_automation_level()
    if level == None:
        self.userinterface.radiobutton_updates_inst_sec.setChecked(False)
        self.userinterface.radiobutton_updates_download.setChecked(False)
        self.userinterface.radiobutton_updates_notify.setChecked(False)
    if level == softwareproperties.UPDATE_MANUAL or\
       level == softwareproperties.UPDATE_NOTIFY:
        self.userinterface.radiobutton_updates_notify.setChecked(True)
    elif level == softwareproperties.UPDATE_DOWNLOAD:
        self.userinterface.radiobutton_updates_download.setChecked(True)
    elif level == softwareproperties.UPDATE_INST_SEC:
        self.userinterface.radiobutton_updates_inst_sec.setChecked(True)

  def init_distro(self):
    text = _("Software updates")
    self.userinterface.groupBox_updates.setTitle(text)
    text = _("Ubuntu Software")
    self.userinterface.tabWidget.setTabText(0, text)

    # Setup the checkbuttons for the components
    for child in self.userinterface.vbox_dist_comps_frame.children():
        if isinstance(child, QGridLayout):
            self.vbox_dist_comps = child
        elif isinstance(child, QObject):
            print("passing")
        else:
            print("child: " + str(type(child)))
            child.hide()
            del child
    self.checkboxComps = {}
    for comp in self.distro.source_template.components:
        # TRANSLATORS: Label for the components in the Internet section
        #              first %s is the description of the component
        #              second %s is the code name of the comp, eg main, universe
        label = _("%s (%s)") % (comp.get_description(), comp.name)
        checkbox = QCheckBox(label, self.userinterface.vbox_dist_comps_frame)
        checkbox.setObjectName(comp.name)
        self.checkboxComps[checkbox] = comp

        # setup the callback and show the checkbutton
        checkbox.clicked.connect(self.on_component_toggled)
        self.vbox_dist_comps.addWidget(checkbox)

    # Setup the checkbuttons for the child repos / updates
    for child in self.userinterface.vbox_updates_frame.children():
        if isinstance(child, QGridLayout):
            self.vbox_updates = child
        else:
            del child
    if len(self.distro.source_template.children) < 1:
        self.userinterface.vbox_updates_frame.hide()
    self.checkboxTemplates = {}
    for template in self.distro.source_template.children:
        # Do not show source entries in there
        if template.type == "deb-src":
            continue

        checkbox = QCheckBox(template.description,
                             self.userinterface.vbox_updates_frame)
        checkbox.setObjectName(template.name)
        self.checkboxTemplates[checkbox] = template
        # setup the callback and show the checkbutton
        checkbox.clicked.connect(self.on_checkbutton_child_toggled)
        self.vbox_updates.addWidget(checkbox)

    # If no components are enabled there will be no need for updates
    if len(self.distro.enabled_comps) < 1:
        self.userinterface.vbox_updates_frame.setEnabled(False)
    else:
        self.userinterface.vbox_updates_frame.setEnabled(True)

    self.init_mirrors()
    self.userinterface.combobox_server.activated.connect(self.on_combobox_server_changed)

  def init_mirrors(self):
    # Intiate the combobox which allows the user to specify a server for all
    # distro related sources
    seen_server_new = []
    self.mirror_urls = []
    self.userinterface.combobox_server.clear()
    for (name, uri, active) in self.distro.get_server_list():
        ##server_store.append([name, uri, False])
        self.userinterface.combobox_server.addItem(name)
        self.mirror_urls.append(uri)
        if [name, uri] in self.seen_server:
            self.seen_server.remove([name, uri])
        elif uri != None:
            seen_server_new.append([name, uri])
        if active == True:
            self.active_server = self.userinterface.combobox_server.count() - 1
            self.userinterface.combobox_server.setCurrentIndex(self.userinterface.combobox_server.count() - 1)
    for [name, uri] in self.seen_server:
        self.userinterface.combobox_server.addItem(name)
        self.mirror_urls.append(uri)
    self.seen_server = seen_server_new
    # add a separator and the option to choose another mirror from the list
    ##FIXME server_store.append(["sep", None, True])
    self.userinterface.combobox_server.addItem(_("Other..."))
    self.other_mirrors_index = self.userinterface.combobox_server.count() - 1

  def init_apt_cache(self):
    apt_pkg.init_config()
    apt_pkg.init_system()
    self.apt_cache = apt_pkg.Cache(None)
    self.depcache = apt_pkg.DepCache(self.apt_cache)
    self.records = apt_pkg.PackageRecords(self.apt_cache)

  def update_apt_cache(self):
    self.apt_cache = apt_pkg.Cache(None)
    self.depcache = apt_pkg.DepCache(self.apt_cache)
    self.records = apt_pkg.PackageRecords(self.apt_cache)

  def show_distro(self):
    """
    Represent the distro information in the user interface
    """
    # Enable or disable the child source checkbuttons
    for child in self.userinterface.vbox_updates_frame.children():
        if isinstance(child, QCheckBox):
            template = self.checkboxTemplates[child]
            (active, inconsistent) = self.get_comp_child_state(template)
            if inconsistent:
                child.setCheckState(Qt.PartiallyChecked)
            elif active:
                child.setCheckState(Qt.Checked)
            else:
                child.setCheckState(Qt.Unchecked)

    # Enable or disable the component checkbuttons
    for child in self.userinterface.vbox_dist_comps_frame.children():
        if isinstance(child, QCheckBox):
            template = self.checkboxComps[child]
            (active, inconsistent) = self.get_comp_download_state(template)
            if inconsistent:
                child.setCheckState(Qt.PartiallyChecked)
            elif active:
                child.setCheckState(Qt.Checked)
            else:
                child.setCheckState(Qt.Unchecked)

    # If no components are enabled there will be no need for updates
    # and source code
    if len(self.distro.enabled_comps) < 1:
        self.userinterface.vbox_updates_frame.setEnabled(False)
        self.userinterface.checkbutton_source_code.setEnabled(False)
    else:
        self.userinterface.vbox_updates_frame.setEnabled(True)
        self.userinterface.checkbutton_source_code.setEnabled(True)

    # Check for source code sources
    source_code_state = self.get_source_code_state()
    if source_code_state == True:
        self.userinterface.checkbutton_source_code.setCheckState(Qt.Checked)
    elif source_code_state == None:
        self.userinterface.checkbutton_source_code.setCheckState(Qt.PartiallyChecked)
    else:
        self.userinterface.checkbutton_source_code.setCheckState(Qt.Unchecked)

    # Will show a short explanation if no CDROMs are used
    if len(self.get_cdrom_sources()) == 0:
        self.userinterface.treeview_cdroms.hide()
        self.userinterface.textview_no_cd.show()
        self.userinterface.groupBox_cdrom.hide()
    else:
        self.userinterface.treeview_cdroms.show()
        self.userinterface.textview_no_cd.hide()
        self.userinterface.groupBox_cdrom.show()

    # Output a lot of debug stuff
    if self.options.debug == True or self.options.massive_debug == True:
        print("ENABLED COMPS: %s" % self.distro.enabled_comps)
        print("INTERNET COMPS: %s" % self.distro.download_comps)
        print("MAIN SOURCES")
        for source in self.distro.main_sources:
            self.print_source_entry(source)
        print("CHILD SOURCES")
        for source in self.distro.child_sources:
            self.print_source_entry(source)
        print("CDROM SOURCES")
        for source in self.distro.cdrom_sources:
            self.print_source_entry(source)
        print("SOURCE CODE SOURCES")
        for source in self.distro.source_code_sources:
            self.print_source_entry(source)
        print("DISABLED SOURCES")
        for source in self.distro.disabled_sources:
            self.print_source_entry(source)
        print("ISV")
        for source in self.sourceslist_visible:
            self.print_source_entry(source)

  def set_update_automation_level(self, selected):
    """Call the backend to set the update automation level to the given
       value"""
    if self.userinterface.radiobutton_updates_download.isChecked():
        SoftwareProperties.set_update_automation_level(self, softwareproperties.UPDATE_DOWNLOAD)
    elif self.userinterface.radiobutton_updates_inst_sec.isChecked():
        SoftwareProperties.set_update_automation_level(self, softwareproperties.UPDATE_INST_SEC)
    elif self.userinterface.radiobutton_updates_notify.isChecked():
        SoftwareProperties.set_update_automation_level(self, softwareproperties.UPDATE_NOTIFY)
    self.set_modified_config()

  def on_combobox_release_upgrades_changed(self, combobox):
    """ set the release upgrades policy """
    i = self.userinterface.combobox_release_upgrades.currentIndex()

    self.set_release_upgrades_policy(i)

  def on_combobox_server_changed(self, combobox):
    """
    Replace the servers used by the main and update sources with
    the selected one
    """
    combobox = self.userinterface.combobox_server
    index = combobox.currentIndex()
    if index == self.active_server:
        return
    elif index == self.other_mirrors_index:
        dialogue = DialogMirror(self.userinterface, self.datadir, self.distro, self.custom_mirrors)
        res = dialogue.run()
        if res != None:
            self.distro.change_server(res)
            self.set_modified_sourceslist()
            self.init_mirrors() # update combobox
        else:
            combobox.setCurrentIndex(self.active_server)
    else:
        uri = self.mirror_urls[index]
        if uri != None and len(self.distro.used_servers) > 0:
            self.active_server = index
            self.distro.change_server(uri)
            self.set_modified_sourceslist()
        else:
            self.distro.default_server = uri

  def on_component_toggled(self):
    """
    Sync the components of all main sources (excluding cdroms),
    child sources and source code sources
    """
    # FIXME: I find it rather questionable whether sender will work with pyqt doing weird signal handling
    tickBox = kapp.sender()
    if tickBox.checkState() == Qt.Checked:
        self.enable_component(str(tickBox.objectName()))
    elif tickBox.checkState() == Qt.Unchecked:
        self.disable_component(str(tickBox.objectName()))
    self.set_modified_sourceslist()

    # no way to set back to mixed state so turn off tristate
    tickBox.setTristate(False)

  def on_checkbutton_child_toggled(self):
    """
    Enable or disable a child repo of the distribution main repository
    """
    tickBox = kapp.sender()
    name = str(tickBox.objectName())

    if tickBox.checkState() == Qt.Checked:
        self.enable_child_source(name)
    elif tickBox.checkState() == Qt.Unchecked:
        self.disable_child_source(name)
    # no way to set back to mixed state so turn off tristate
    tickBox.setTristate(False)

  def on_checkbutton_source_code_toggled(self):
    """ Disable or enable the source code for all sources """
    if self.userinterface.checkbutton_source_code.checkState() == Qt.Checked:
        self.enable_source_code_sources()
    elif self.userinterface.checkbutton_source_code.checkState() == Qt.Unchecked:
        self.disable_source_code_sources()
    self.userinterface.checkbutton_source_code.setTristate(False)

  def on_checkbutton_popcon_toggled(self, state):
    """ The user clicked on the popcon paritipcation button """
    self.set_popcon_pariticipation(state)

  def init_isv_sources(self):
    """Read all valid sources into our ListStore"""
    """##FIXME
    # drag and drop support for sources.list
    self.treeview_sources.drag_dest_set(gtk.DEST_DEFAULT_ALL, \
                                        [('text/uri-list',0, 0)], \
                                        gtk.gdk.ACTION_COPY)
    self.treeview_sources.connect("drag_data_received",\
                                  self.on_sources_drag_data_received)
    """

  def on_isv_source_activate(self, treeview, path, column):
    """Open the edit dialog if a channel was double clicked"""
    ##FIXME TODO
    ##self.on_edit_clicked(treeview)

  def on_treeview_sources_cursor_changed(self, item, column):
    """set the sensitiveness of the edit and remove button
       corresponding to the selected channel"""
    # allow to remove the selected channel
    self.userinterface.button_remove.setEnabled(True)
    # disable editing of cdrom sources
    index = self.userinterface.treeview_sources.indexOfTopLevelItem(item)
    source_entry = self.isv_sources[index]
    if source_entry.uri.startswith("cdrom:"):
        self.userinterface.button_edit.setEnabled(False)
    else:
        self.userinterface.button_edit.setEnabled(True)

  def on_cdrom_source_toggled(self, item):
    """Enable or disable the selected channel"""
    self.toggle_source(self.cdrom_sources, self.userinterface.treeview_cdroms, item)

  def on_isv_source_toggled(self, item):
    """Enable or disable the selected channel"""
    self.toggle_source(self.isv_sources, self.userinterface.treeview_sources, item)

  def toggle_source(self, sources, treeview, item):
    """Enable or disable the selected channel"""
    index = treeview.indexOfTopLevelItem(item)
    source_entry = sources[index]
    self.toggle_source_use(source_entry)
    item = treeview.topLevelItem(index) # old item was destroyed
    if item != 0:
      treeview.setCurrentItem(item) # reselect entry after refresh
    else:
      self.userinterface.button_remove.setEnabled(False)
      self.userinterface.button_edit.setEnabled(False)

  def init_keys(self):
    """Setup the user interface parts needed for the key handling"""
    self.userinterface.treeview2.setColumnCount(1)

  def on_treeview_keys_cursor_changed(self, item, column):
    """set the sensitiveness of the edit and remove button
       corresponding to the selected channel"""
    # allow to remove the selected channel
    self.userinterface.button_auth_remove.setEnabled(True)
    # disable editing of cdrom sources
    index = self.userinterface.treeview2.indexOfTopLevelItem(item)

  #FIXME revert automation settings too
  def on_button_revert_clicked(self):
    """Restore the source list from the startup of the dialog"""
    SoftwareProperties.revert(self)
    self.set_modified_sourceslist()
    self.show_auto_update_level()
    self.userinterface.buttonBox.button(QDialogButtonBox.Reset).setEnabled(False)
    self.modified_sourceslist = False

  def set_modified_config(self):
    """The config was changed and now needs to be saved and reloaded"""
    SoftwareProperties.set_modified_config(self)
    self.userinterface.buttonBox.button(QDialogButtonBox.Reset).setEnabled(True)

  def set_modified_sourceslist(self):
    """The sources list was changed and now needs to be saved and reloaded"""
    SoftwareProperties.set_modified_sourceslist(self)
    self.show_distro()
    self.show_isv_sources()
    self.show_cdrom_sources()
    self.userinterface.buttonBox.button(QDialogButtonBox.Reset).setEnabled(True)

  def show_isv_sources(self):
    """ Show the repositories of independent software vendors in the
        third-party software tree view """
    self.isv_sources = self.get_isv_sources()[:]
    self.show_sources(self.isv_sources, self.userinterface.treeview_sources)

    if not self.isv_sources or len(self.isv_sources) < 1:
        self.userinterface.button_remove.setEnabled(False)
        self.userinterface.button_edit.setEnabled(False)

  def show_cdrom_sources(self):
    """ Show CD-ROM/DVD based repositories of the currently used distro in
        the CDROM based sources list """
    self.cdrom_sources = self.get_cdrom_sources()[:]
    self.show_sources(self.cdrom_sources, self.userinterface.treeview_cdroms)

  def show_sources(self, sources, treeview):
    # this workaround purposely replaces treeview.clear() (LP #102792)
    while treeview.topLevelItemCount() > 0:
      treeview.takeTopLevelItem(0)

    items = []
    for source in sources:
        contents = self.render_source(source)
        contents = contents.replace("<b>", "")
        contents = contents.replace("</b>", "")
        item = QTreeWidgetItem([contents])
        if not source.disabled:
            item.setCheckState(0, Qt.Checked)
        else:
            item.setCheckState(0, Qt.Unchecked)
        items.append(item)
    treeview.addTopLevelItems(items)

  def show_keys(self):
    self.userinterface.treeview2.clear()
    for key in self.apt_key.list():
      self.userinterface.treeview2.addTopLevelItem(QTreeWidgetItem([key]))

    if self.userinterface.treeview_sources.topLevelItemCount() < 1:
      self.userinterface.button_auth_remove.setEnabled(False)

  def on_combobox_update_interval_changed(self, index):
    #FIXME! move to backend
    if index != -1:
        value = self.combobox_interval_mapping[index]
        # Only write the key if it has changed
        if not value == apt_pkg.config.find_i(softwareproperties.CONF_MAP["autoupdate"]):
            apt_pkg.config.set(softwareproperties.CONF_MAP["autoupdate"], str(value))
            self.write_config()

  def on_auto_update_toggled(self):
    """Enable or disable automatic updates and modify the user interface
       accordingly"""
    if self.userinterface.checkbutton_auto_update.checkState() == Qt.Checked:
      self.userinterface.combobox_update_interval.setEnabled(True)
      self.userinterface.radiobutton_updates_inst_sec.setEnabled(True)
      self.userinterface.radiobutton_updates_download.setEnabled(True)
      self.userinterface.radiobutton_updates_notify.setEnabled(True)
      # if no frequency was specified use daily
      i = self.userinterface.combobox_update_interval.currentIndex()
      if i == -1:
          i = 0
          self.userinterface.combobox_update_interval.setCurrentIndex(i)
      value = self.combobox_interval_mapping[i]
    else:
      self.userinterface.combobox_update_interval.setEnabled(False)
      self.userinterface.radiobutton_updates_inst_sec.setEnabled(False)
      self.userinterface.radiobutton_updates_download.setEnabled(False)
      self.userinterface.radiobutton_updates_notify.setEnabled(False)
      SoftwareProperties.set_update_automation_level(self, None)
      value = 0
    self.set_update_interval(str(value))

  def on_add_clicked(self):
    """Show a dialog that allows to enter the apt line of a to be used repo"""
    dialog = DialogAdd(self.userinterface, self.sourceslist,
                       self.datadir, self.distro)
    line = dialog.run()
    if line != None:
      self.add_source_from_line(line)
      self.set_modified_sourceslist()

  def on_edit_clicked(self):
    """Show a dialog to edit an ISV source"""
    item = self.userinterface.treeview_sources.currentItem()
    if item is not None:
        index = self.userinterface.treeview_sources.indexOfTopLevelItem(item)
        dialogue = DialogEdit(self.userinterface, self.sourceslist, self.isv_sources[index], self.datadir)
        result = dialogue.run()
        if result == QDialog.Accepted:
            self.set_modified_sourceslist()
            self.show_isv_sources()

  # FIXME:outstanding from merge
  def on_isv_source_activated(self, treeview, path, column):
     """Open the edit dialog if a channel was double clicked"""
     ##FIXME TODO
     # check if the channel can be edited
     if self.button_edit.get_property("sensitive") == True:
         self.on_edit_clicked(treeview)

  def on_remove_clicked(self):
    """Remove the selected source"""
    item = self.userinterface.treeview_sources.currentItem()
    if item is not None:
        index = self.userinterface.treeview_sources.indexOfTopLevelItem(item)
        self.remove_source(self.isv_sources[index])
        self.show_isv_sources()

  def add_key_clicked(self):
    """Provide a file chooser that allows to add the gnupg of a trusted
       software vendor"""
    home = QDir.homePath()
    if "SUDO_USER" in os.environ:
        home = os.path.expanduser("~%s" % os.environ["SUDO_USER"])
    url = KUrl.fromPath(home)
    filename = KFileDialog.getOpenFileName(url, 'application/pgp-keys', self.userinterface, _("Import key"))
    if filename:
      if not self.add_key(filename):
        title = _("Error importing selected file")
        text = _("The selected file may not be a GPG key file " \
                "or it might be corrupt.")
        #KMessageBox.sorry(self.userinterface, text, title)
        QMessageBox.warning(self.userinterface, title, text)
      self.show_keys()

  def remove_key_clicked(self):
    """Remove a trusted software vendor key"""
    item = self.userinterface.treeview2.currentItem()
    if item == None:
        return
    key = item.text(0)
    if not self.remove_key(key[:16]):
      title = _("Error removing the key")
      text = _("The key you selected could not be removed. "
               "Please report this as a bug.")
      #KMessageBox.sorry(self.userinterface, text, title)
      QMessageBox.warning(self.userinterface, title, text)
    self.show_keys()

  def on_restore_clicked(self):
    """Restore the original keys"""
    self.update_apt_cache()
    self.show_keys()

  def on_pktask_progress(self, progress, ptype, udata=(None,)):
    if ptype == packagekit.ProgressType.PERCENTAGE:
      perc = progress.get_property('percentage')
      self._pdialog.setValue(perc)

  def on_pktask_finish(self, source, result, udata=(None,)):
    results = None
    try:
        results = self._pktask.generic_finish(result)
    except Exception as e:
      QMessageBox.warning(self.userinterface, _("Could not refresh cache"), str(e))
    self._pdialog.hide()
    kapp.quit()

  def on_close_button(self):
    """Show a dialog that a reload of the channel information is required
       only if there is no parent defined"""
    if self.modified_sourceslist == True and self.options.no_update == False:
        messageBox = QMessageBox(self.userinterface)
        messageBox.setIcon(QMessageBox.Information)
        reloadButton = messageBox.addButton(_("Reload"), QMessageBox.AcceptRole)
        messageBox.addButton(QMessageBox.Close)
        primary = _("Your local copy of the software catalog is out of date.")
        secondary = _("A new copy will be downloaded.")
        text = "%s<br /><br /><small>%s</small>" % (primary, secondary)
        messageBox.setText(text)
        messageBox.exec_()
        if (messageBox.clickedButton() == reloadButton):
                self._pdialog = QProgressDialog("<big>{}</big>".format(_("Refreshing software cache")),
                                                "Cancel", 0, 100, self.userinterface)
                self._pdialog.setWindowTitle(_("Cache Refresh"))
                self._pdialog.setCancelButton(None)
                self._pdialog.setAutoClose(False)
                self._pdialog.setMinimumWidth(210)
                self._pdialog.setMinimumHeight(60)

                self._pktask = packagekit.Task()
                self._pdialog.show()
                self.userinterface.hide()
                try:
                    self._pktask.refresh_cache_async (False, # force
                                                  None,  # GCancellable
                                                  self.on_pktask_progress,
                                                  (None,), # user data
                                                  self.on_pktask_finish,
                                                  (None,));
                except Exception as e:
                    print("Error while requesting cache refresh: {}".format (e))
        else:
            # refresh not wanted, quit immediately
            kapp.quit()
    else:
        # no changes, no cache refresh needed.
        # we can quit.
        kapp.quit()


  def on_button_add_cdrom_clicked(self):
    '''Show a dialog that allows to add a repository located on a CDROM
       or DVD'''
    # testing
    #apt_pkg.config.set("APT::CDROM::Rename","true")

    saved_entry = apt_pkg.config.find("Dir::Etc::sourcelist")
    tmp = tempfile.NamedTemporaryFile()
    apt_pkg.config.set("Dir::Etc::sourcelist",tmp.name)
    progress = CdromProgress(self.datadir, self, kapp)
    cdrom = apt_pkg.Cdrom()
    # if nothing was found just return
    try:
      res = cdrom.add(progress)
    except SystemError as msg:
      title = _("CD Error")
      primary = _("Error scanning the CD")
      text = "%s\n\n<small>%s</small>" % (primary, msg)
      #KMessageBox.sorry(self.userinterface, text, title)
      QMessageBox.warning(self.userinterface, title, text)
      return
    finally:
      apt_pkg.config.set("Dir::Etc::sourcelist",saved_entry)
      progress.close()

    if res == False:
      return
    # read tmp file with source name (read only last line)
    line = ""
    for x in open(tmp.name):
      line = x
    if line != "":
      full_path = "%s%s" % (apt_pkg.config.find_dir("Dir::Etc"),saved_entry)
      # insert cdrom source first, so that it has precedence over network sources
      self.sourceslist.list.insert(0, SourceEntry(line,full_path))
      self.set_modified_sourceslist()

  def run(self):
    kapp.exec_()

  def on_driver_changes_progress(self, progress, ptype, data=None):
    self.button_driver_revert.setVisible(False)
    self.button_driver_apply.setVisible(False)
    self.button_driver_restart.setVisible(False)
    self.button_driver_cancel.setVisible(True)
    self.progress_bar.setVisible(True)

    self.userinterface.label_driver_action.setText("Applying changes...")
    if ptype == packagekit.ProgressType.PERCENTAGE:
        prog_value = progress.get_property('percentage')
        self.progress_bar.setValue(prog_value)

  def on_driver_changes_finish(self, source, result, installs_pending):
    results = None
    try:
      results = self.pk_task.generic_finish(result)
    except Exception as e:
      self.on_driver_changes_revert()
      QMessageBox.warning(self.userinterface, _("Error while applying changes"), str(e))

    if not installs_pending:
      self.progress_bar.setVisible(False)
      self.clear_changes()
      self.update_apt_cache()
      self.set_driver_action_status()
      self.update_label_and_icons_from_status()
      self.button_driver_revert.setVisible(True)
      self.button_driver_apply.setVisible(True)
      self.button_driver_cancel.setVisible(False)
    #self.scrolled_window_drivers.set_sensitive(True)

  def on_driver_changes_error(self, transaction, error_code, error_details):
    self.on_driver_changes_revert()
    self.set_driver_action_status()
    self.update_label_and_icons_from_status()
    self.button_driver_revert.setVisible(True)
    self.button_driver_apply.setVisible(True)
    self.button_driver_cancel.setVisible(False)
    #self.scrolled_window_drivers.set_sensitive(True)

  def on_driver_changes_cancellable_changed(self, transaction, cancellable):
    self.button_driver_cancel.setEnabled(cancellable)

  def on_driver_changes_apply(self, button):
    self.pk_task = packagekit.Task()
    button = self.userinterface.sender()
    installs = []
    removals = []

    has_nvidia = False
    for pkg in self.driver_changes:
      if pkg.current_ver:
          removals.append(self.get_package_id(self.apt_cache, pkg))
          # The main NVIDIA package is only a metapackage.
          # We need to collect its dependencies, so that
          # we can uninstall the driver properly.
          if 'nvidia' in pkg.name:
            has_nvidia = True
            for dep in self.get_dependencies(self.apt_cache, pkg, 'nvidia'):
              if dep.current_ver:
                removals.append(self.get_package_id(self.apt_cache, dep))
      else:
        installs.append(self.get_package_id(self.apt_cache, pkg))

    self.cancellable = Gio.Cancellable()
    try:
      if removals:
        installs_pending = False
        if installs:
            installs_pending = True
        self.pk_task.remove_packages_async(removals,
                  False,  # allow deps
                  True,  # autoremove
                  self.cancellable,  # cancellable
                  self.on_driver_changes_progress,
                  (None, ),  # progress data
                  self.on_driver_changes_finish,  # callback ready
                  installs_pending  # callback data
         )
      if installs:
        if has_nvidia:
          to_install = []
          for item in installs:
            name = item.split(';')[0]
            to_install.append(name)
        self.pk_task.install_packages_async(installs,
              self.cancellable,  # cancellable
              self.on_driver_changes_progress,
              (None, ),  # progress data
              self.on_driver_changes_finish,  # GAsyncReadyCallback
              False  # ready data
       )
      self.button_driver_revert.setEnabled(False)
      self.button_driver_apply.setEnabled(False)
      #self.scrolled_window_drivers.set_sensitive(False)
    except Exception as e:
      print("Warning: install transaction not completed successfully: {}".format(e))

  def on_driver_changes_revert(self, button_revert=None):
    # HACK: set all the "Do not use" first; then go through the list of the
    #       actually selected drivers.
    for button in self.no_drv:
      button.setChecked(True)

    for alias in self.orig_selection:
      button = self.orig_selection[alias]
      button.setChecked(True)

    self.clear_changes()

    self.button_driver_revert.setEnabled(False)
    self.button_driver_apply.setEnabled(False)

  def on_driver_changes_cancel(self, button_cancel):
    self.transaction.cancel()
    self.clear_changes()

  def on_driver_restart_clicked(self, button_restart):
    if 'XDG_CURRENT_DESKTOP' in os.environ:
      desktop = os.environ['XDG_CURRENT_DESKTOP']
    else:
      desktop = 'Unknown'

    if (desktop == 'ubuntu:GNOME' and os.path.exists('/usr/bin/gnome-session-quit')):
      # argument presents a dialog to cancel reboot
      subprocess.call(['gnome-session-quit', '--reboot'])
    elif (desktop == 'XFCE' and
      os.path.exists('/usr/bin/xfce4-session-logout')):
      subprocess.call(['xfce4-session-logout'])
    elif (desktop == 'LXDE' and os.path.exists('/usr/bin/lubuntu-logout')):
      subprocess.call(['lubuntu-logout'])
    elif (desktop == 'LXQt' and os.path.exists('/usr/bin/lxqt-leave')):
      subprocess.call(['lxqt-leave'])

  def clear_changes(self):
    self.orig_selection = {}
    self.driver_changes = []

  def init_drivers(self):
    """Additional Drivers tab"""
    self.button_driver_revert = QPushButton("Revert")
    self.button_driver_apply = QPushButton("Apply Changes")
    self.button_driver_cancel = QPushButton("Cancel")
    self.button_driver_restart = QPushButton("Restart...")

    self.button_driver_revert.clicked.connect(self.on_driver_changes_revert)
    self.button_driver_apply.clicked.connect(self.on_driver_changes_apply)
    self.button_driver_cancel.clicked.connect(self.on_driver_changes_cancel)
    self.button_driver_restart.clicked.connect(self.on_driver_restart_clicked)

    self.button_driver_revert.setEnabled(False)
    self.button_driver_revert.setVisible(True)
    self.button_driver_apply.setEnabled(False)
    self.button_driver_apply.setVisible(True)
    self.button_driver_cancel.setVisible(False)
    self.button_driver_restart.setVisible(False)

    #self.userinterface.box_driver_action.addWidget(self.userinterface.label_driver_action)
    self.userinterface.box_driver_action.addStretch()
    self.userinterface.box_driver_action.addWidget(self.button_driver_apply)
    self.userinterface.box_driver_action.addWidget(self.button_driver_revert)
    self.userinterface.box_driver_action.addWidget(self.button_driver_restart)
    self.userinterface.box_driver_action.addWidget(self.button_driver_cancel)

    self.label_driver_detail = QLabel("Searching for available drivers...")
    self.label_driver_detail.setAlignment(Qt.AlignCenter)
    self.userinterface.box_driver_detail.addWidget(self.label_driver_detail)

    self.progress_bar = QProgressBar()

    self.userinterface.box_driver_action.addWidget(self.progress_bar)
    self.progress_bar.setVisible(False)

    self.devices = {}
    self.detect_called = False
    self.driver_changes = []
    self.orig_selection = {}
        # HACK: the case where the selection is actually "Do not use"; is a little
        #       tricky to implement because you can't check for whether a package is
        #       installed or any such thing. So let's keep a list of all the
        #       "Do not use" radios, set those active first, then iterate through
        #       orig_selection when doing a Reset.
    self.no_drv = []
    self.nonfree_drivers = 0
    self.ui_building = False

  def detect_drivers(self):
    # WARNING: This is run in a separate thread.
    self.detect_called = True
    self.init_apt_cache()
    try:
      self.devices = detect.system_device_drivers(self.apt_cache)
    except:
      # Catch all exceptions and feed them to apport.
      #GLib.idle_add(self.label_driver_detail.set_text, _("An error occurred while searching for drivers."))
      self.label_driver_detail.setText("An error occurred while searching for drivers.")
      # For apport to catch this exception. See
      # http://bugs.python.org/issue1230540
      sys.excepthook(*sys.exc_info())
      return
  def on_driver_selection_changed(self, modalias, pkg_name=None):
    button = self.userinterface.sender()
    #print(modalias)
    #print(pkg_name)

    if self.ui_building:
      return

    pkg = None

    if pkg_name:
      pkg = self.apt_cache[pkg_name]
      # Add the matching linux modules package when available
      try:
        modules_package = detect.get_linux_modules_metapackage(self.apt_cache, pkg_name)
        modules_package_obj = self.apt_cache[modules_package]
        if modules_package and not modules_package_obj.current_ver:
          self.driver_changes.append(modules_package_obj)
      except (KeyError, TypeError):
          pass

    if button.isChecked():
      if pkg in self.driver_changes:
        self.driver_changes.remove(pkg)

      if (pkg is not None and modalias in self.orig_selection and button is not self.orig_selection[modalias]):
        self.driver_changes.append(pkg)
    else:
      if pkg in self.driver_changes:
        self.driver_changes.remove(pkg)

      # for revert; to re-activate the original radio buttons.
      if modalias not in self.orig_selection:
          self.orig_selection[modalias] = button

      if (pkg is not None and pkg not in self.driver_changes and pkg.current_ver):
          self.driver_changes.append(pkg)

    self.button_driver_revert.setEnabled(bool(self.driver_changes))
    self.button_driver_apply.setEnabled(bool(self.driver_changes))

  def gather_device_data(self, device):
    '''Get various device data used to build the GUI.

          return a tuple of (overall_status string, icon, drivers dict).
          the drivers dict is using this form:
            {"recommended/alternative": {pkg_name: {
                                                      'selected': True/False
                                                      'description':
'description'
                                                      'builtin': True/False
                                                    }
                                         }}
             "manually_installed": {"manual": {'selected': True, 'description':
description_string}}
             "no_driver": {"no_driver": {'selected': True/False, 'description':
description_string}}

             Please note that either manually_installed and no_driver are set to
None if not applicable
             (no_driver isn't present if there are builtins)
        '''

    possible_overall_status = {
            'recommended': (_("This device is using the recommended driver."),
"recommended-driver"),
            'alternative': (_("This device is using an alternative driver."),
"other-driver"),
            'manually_installed': (_("This device is using a manually-installed driver."), "other-driver"),
            'no_driver': (_("This device is not working."), "disable-device")
        }

    returned_drivers = {'recommended': {}, 'alternative': {},
'manually_installed': {}, 'no_driver': {}}
    have_builtin = False
    one_selected = False
    try:
      if device['manual_install']:
        returned_drivers['manually_installed'] = {True: {'selected': True,'description': _("Continue using a manually installed driver")}}
    except KeyError:
      pass

    for pkg_driver_name in device['drivers']:
      current_driver = device['drivers'][pkg_driver_name]

      # get general status
      driver_status = 'alternative'
      try:
        if current_driver['recommended'] and current_driver['from_distro']:
          driver_status = 'recommended'
      except KeyError:
        pass

      builtin = False
      try:
        if current_driver['builtin']:
          builtin = True
          have_builtin = True
      except KeyError:
        pass

      try:
        pkg = self.apt_cache[pkg_driver_name]
        installed = pkg.current_ver or False
        candidate = self.depcache.get_candidate_ver(pkg)
        if candidate is not None:
          self.records.lookup(candidate.file_list[0])
          description = _("Using {} from {}").format(self.records.short_desc, pkg_driver_name)
        else:
          description = _("Using {}").format(pkg_driver_name)
      except KeyError:
        print("WARNING: a driver ({}) doesn't have any available package associated: {}".format(pkg_driver_name, current_driver))
        continue

      # gather driver description
      if current_driver['free']:
        licence = _("open source")
      else:
        licence = _("proprietary")

      if driver_status == 'recommended':
        base_string = _("{base_description} ({licence}, tested)")
      else:
        base_string = _("{base_description} ({licence})")
      description = base_string.format(base_description=description,
licence=licence)

      selected = False
      if not builtin and not returned_drivers['manually_installed']:
        selected = installed
        if installed:
          selected = True
          one_selected = True

      returned_drivers[driver_status].setdefault(pkg_driver_name, {'selected': selected,'description': description,'builtin': builtin})

    # adjust making the needed addition
    if not have_builtin:
      selected = False
      if not one_selected:
        selected = True
      returned_drivers["no_driver"] = {True: {'selected': selected, 'description': _("Do not use the device")}}
    else:
      # we have a builtin and no selection: builtin is the selected one then
      if not one_selected:
        for section in ('recommended', 'alternative'):
          for pkg_name in returned_drivers[section]:
            if returned_drivers[section][pkg_name]['builtin']:
              returned_drivers[section][pkg_name]['selected'] = True

    # compute overall status
    for section in returned_drivers:
      for keys in returned_drivers[section]:
        if returned_drivers[section][keys]['selected']:
          (overall_status, icon) = possible_overall_status[section]

    return (overall_status, icon, returned_drivers)

  def show_drivers(self):
    if not self.devices:
      # No drivers found.
      self.label_driver_detail.setText("No additional drivers available.")
      return
    else:
      self.label_driver_detail.hide()

    self.option_group = {}
    self.radio_button = {}
    self.ui_building = True
    self.dynamic_device_status = {}
    for device in sorted(self.devices.keys()):
      (overall_status, icon, drivers) = self.gather_device_data(self.devices[device])

      driver_status = QLabel()
      driver_status.setAlignment(Qt.AlignTop)
      driver_status.setAlignment(Qt.AlignHCenter)
      pixmap = QIcon.fromTheme(icon).pixmap(QSize(16, 16))
      driver_status.setPixmap(pixmap)

      device_box = QHBoxLayout()
      device_box.addWidget(driver_status)
      device_detail = QVBoxLayout()
      device_box.addLayout(device_detail,1)#1 for priority over the icon to stretch

      widget = QLabel("{}: {}".format(self.devices[device].get('vendor', _('Unknown')), self.devices[device].get('model', _('Unknown'))))
      widget.setAlignment(Qt.AlignLeft)
      device_detail.addWidget(widget)
      widget = QLabel("<small>{}</small>".format(overall_status))
      widget.setAlignment(Qt.AlignLeft)
      #widget.set_use_markup(True)
      device_detail.addWidget(widget)
      self.dynamic_device_status[device] = (driver_status, widget)

      self.option_group[device] = None
      # define the order of introspection
      for section in ('recommended', 'alternative', 'manually_installed', 'no_driver'):
        for driver in drivers[section]:
          self.radio_button = QRadioButton(drivers[section][driver]['description'])
          if self.option_group[device]:
            self.option_group[device].addButton(self.radio_button)
          else:
            self.option_group[device] = QButtonGroup()
            self.option_group[device].addButton(self.radio_button)

          device_detail.addWidget(self.radio_button)
          self.radio_button.setChecked(drivers[section][driver]['selected'])

          if section == 'no_driver':
            self.no_drv.append(self.radio_button)
          if section in ('manually_install', 'no_driver') or ('builtin' in drivers[section][driver] and drivers[section][driver]['builtin']):
            self.radio_button.toggled.connect(partial( self.on_driver_selection_changed, device))
          else:
            self.radio_button.toggled.connect(partial( self.on_driver_selection_changed, device, driver))
          if drivers['manually_installed'] and section != 'manually_installed':
            self.radio_button.setEnabled(False)

      self.userinterface.box_driver_detail.addLayout(device_box)

    self.userinterface.box_driver_detail.addStretch()
    self.ui_building = False
    #self.userinterface.box_driver_detail.show_all()
    self.set_driver_action_status()

  def update_label_and_icons_from_status(self):
    '''Update the current label and icon, computing the new device status'''

    for device in self.devices:
      (overall_status, icon, drivers) = self.gather_device_data(self.devices[device])
      (driver_status, widget) = self.dynamic_device_status[device]

      pixmap = QIcon.fromTheme(icon).pixmap(QSize(16, 16))
      driver_status.setPixmap(pixmap)
      widget.setText("<small>{}</small>".format(overall_status))

  def set_driver_action_status(self):
    # Update the label in case we end up having some kind of proprietary driver in use.
    if (os.path.exists('/var/run/reboot-required')):
      self.userinterface.label_driver_action.setText("You need to restart the computer to complete the driver changes.")
      self.button_driver_restart.setVisible(True)
      return

    self.nonfree_drivers = 0
    for device in self.devices:
      for pkg_name in self.devices[device]['drivers']:
        pkg = self.apt_cache[pkg_name]
        if not self.devices[device]['drivers'][pkg_name]['free'] and pkg.current_ver:
          self.nonfree_drivers = self.nonfree_drivers + 1

    if self.nonfree_drivers > 0:
      text = "%s proprietary driver(s) in use." % ( self.nonfree_drivers)
      self.userinterface.label_driver_action.setText(text)
    else:
      self.userinterface.label_driver_action.setText(_("No proprietary drivers are in use."))



#!/usr/bin/python3
# -*- coding: utf-8 -*-

from __future__ import print_function

from gi.repository import GLib

import apt_pkg
import aptsources.distro
import aptsources.sourceslist

import dbus
import logging
import glob
import os
import subprocess
import sys
import threading
import time
import unittest

from dbus.mainloop.glib import DBusGMainLoop

sys.path.insert(0, "../")
from softwareproperties.dbus.SoftwarePropertiesDBus import (
    SoftwarePropertiesDBus, DBUS_BUS_NAME, DBUS_PATH, DBUS_INTERFACE_NAME)
from softwareproperties import (
    UPDATE_INST_SEC, UPDATE_DOWNLOAD, UPDATE_NOTIFY)

try:
    DPKG_ARCH = subprocess.check_output(
        ["dpkg", "--print-architecture"]).strip().decode("utf-8")
except subprocess.CalledProcessError:
    sys.stderr.write("WARNING: Failed to read dpkg arch")
    DPKG_ARCH = None

if DPKG_ARCH in ("i386", "amd64"):
    PRIMARY_MIRROR = "http://archive.ubuntu.com/ubuntu" 
else:
    PRIMARY_MIRROR = "http://ports.ubuntu.com/ubuntu-ports"


def get_test_source_line():
    distro_release = get_distro_release()
    return "deb %s %s main restricted #"\
           " comment with unicode äöü" % (PRIMARY_MIRROR, distro_release)


def get_dpkg_arch():
    return subprocess.check_output(
        ["dpkg", "--print-architecture"]).strip().decode("utf-8")


def get_distro_release():
    return "bionic"


def clear_apt_config():
    etc_apt = os.path.join(os.path.dirname(__file__), "aptroot", "etc", "apt")
    for dirpath, dirnames, filenames in os.walk(etc_apt):
        for name in filenames:
            path = os.path.join(dirpath, name)
            if os.path.isfile(path):
                os.unlink(path)

    for d in ["apt.conf.d", "sources.list.d", "trusted.gpg.d", "auth.conf.d"]:
        os.makedirs(os.path.join(etc_apt, d), exist_ok=True)

def create_sources_list():
    s = get_test_source_line() + "\n"
    name = os.path.join(os.path.dirname(__file__),
                        "aptroot", "etc", "apt", "sources.list")
    with open(name, "w") as f:
        f.write(s)
    return name


def session_bus_thread():
    DBusGMainLoop(set_as_default=True)
    loop = GLib.MainLoop()
    threading.current_thread().loop = loop

    bus = dbus.SessionBus(private=True)
    bus.set_exit_on_disconnect(False)

    rootdir = os.path.join(os.path.dirname(__file__), "aptroot")
    spd = SoftwarePropertiesDBus(bus, rootdir=rootdir)
    spd.enforce_polkit = False
    loop.run()
    bus.close()


class TestDBus(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        for k in apt_pkg.config.keys():
            apt_pkg.config.clear(k)
        apt_pkg.init()
        clear_apt_config()
        # create sources.list file
        create_sources_list()
        cls.get_distro = aptsources.distro.get_distro
        aptsources.distro.get_distro = lambda *a, **b: aptsources.distro.UbuntuDistribution(id="Ubuntu", codename="bionic", description="18.04", release="bionic")
        cls.thread = threading.Thread(target=session_bus_thread)
        cls.thread.loop = None
        cls.thread.start()
        while not cls.thread.loop or not cls.thread.loop.is_running():
            time.sleep(1)

    @classmethod
    def tearDownClass(cls):
        cls.thread.loop.quit()
        cls.thread.join()
        aptsources.distro.get_distro = cls.get_distro

    def setUp(self):
        # keep track of signal emissions
        self.sources_list_count = 0
        self.distro_release = get_distro_release()
        create_sources_list()
        self._sourceslist = aptsources.sourceslist.SourcesList()
        # create the client proxy
        self.bus = dbus.SessionBus(private=True, mainloop=DBusGMainLoop())
        self.bus.set_exit_on_disconnect(False)
        self.iface = dbus.Interface(self.bus.get_object(DBUS_BUS_NAME, DBUS_PATH),
                                    DBUS_INTERFACE_NAME)
        self._signal_id = self.iface.connect_to_signal(
            "SourcesListModified", self._on_sources_list_modified)

    def tearDown(self):
        # ensure we remove the "modified" signal again
        self._signal_id.remove()
        self.bus.close()

    def _on_sources_list_modified(self):
        #print("_on_modified_sources_list")
        self.sources_list_count += 1

    @property
    def sourceslist(self):
        self._sourceslist.refresh()
        return ''.join([str(e) for e in self._sourceslist])

    @property
    def enabled_sourceslist(self):
        self._sourceslist.refresh()
        return ''.join([str(e) for e in self._sourceslist
                        if not e.invalid and not e.disabled])

    def _debug_sourceslist(self, text=""):
        logging.debug("sourceslist: %s '%s'" % (text, self.sourceslist))

    # this is an async call - give it a few seconds to catch up with what we expect
    def _assert_eventually(self, prop, n):
        for i in range(9):
            if getattr(self, prop) == n:
                self.assertEqual(getattr(self, prop), n)
            else:
                time.sleep(1)
        # nope, you die now
        self.assertEqual(getattr(self, prop), n)

    def test_enable_disable_component(self):
        # ensure its not there
        self.assertNotIn("universe", self.sourceslist)
        # enable
        self.iface.EnableComponent("universe")
        self._debug_sourceslist("2")
        self.assertIn("universe", self.sourceslist)
        # disable again
        self.iface.DisableComponent("universe")
        self._debug_sourceslist("3")
        self.assertNotIn("universe", self.sourceslist)
        self._assert_eventually("sources_list_count", 2)

    def test_enable_enable_disable_source_code_sources(self):
        # ensure its not there
        self._debug_sourceslist("4")
        self.assertNotIn('deb-src', self.enabled_sourceslist)
        # enable
        self.iface.EnableSourceCodeSources()
        self._debug_sourceslist("5")
        self.assertIn('deb-src', self.enabled_sourceslist)
        # disable again
        self.iface.DisableSourceCodeSources()
        self._debug_sourceslist("6")
        self.assertNotIn('deb-src', self.enabled_sourceslist)
        self._assert_eventually("sources_list_count", 3)

    def test_enable_child_source(self):
        child_source = "%s-updates" % self.distro_release
        # ensure its not there
        self._debug_sourceslist("7")
        self.assertNotIn(child_source, self.sourceslist)
        # enable
        self.iface.EnableChildSource(child_source)
        self._debug_sourceslist("8")
        self.assertIn(child_source, self.sourceslist)
        # disable again
        self.iface.DisableChildSource(child_source)
        self._debug_sourceslist("9")
        self.assertNotIn(child_source, self.sourceslist)
        self._assert_eventually("sources_list_count", 2)

    def test_toggle_source(self):
        # test toggle
        source = get_test_source_line()
        self.iface.ToggleSourceUse(source)
        self._debug_sourceslist("10")
        primary_debline = "# deb %s" % PRIMARY_MIRROR
        self.assertIn(primary_debline, self.sourceslist)
        # to disable the line again, we need to match the new "#"
        source = "# " + source
        self.iface.ToggleSourceUse(source)
        self._debug_sourceslist("11")
        self.assertNotIn(primary_debline, self.sourceslist)

        self._assert_eventually("sources_list_count", 2)

    def test_replace_entry(self):
        # test toggle
        source = get_test_source_line()
        source_new = "deb http://xxx/ %s" % self.distro_release
        self.iface.ReplaceSourceEntry(source, source_new)
        self._debug_sourceslist("11")
        self.assertIn(source_new, self.sourceslist)
        self.assertNotIn(source, self.sourceslist)
        self._assert_eventually("sources_list_count", 1)
        self.iface.ReplaceSourceEntry(source_new, source)
        self._assert_eventually("sources_list_count", 2)

    def test_popcon(self):
        # ensure its set to no
        popcon_p = os.path.join(os.path.dirname(__file__),
                              "aptroot", "etc", "popularity-contest.conf")
        with open(popcon_p) as f:
            popcon = f.read()
            self.assertIn('PARTICIPATE="no"', popcon)
        # toggle
        self.iface.SetPopconPariticipation(True)
        with open(popcon_p) as f:
            popcon = f.read()
            self.assertIn('PARTICIPATE="yes"', popcon)
            self.assertNotIn('PARTICIPATE="no"', popcon)
        # and back
        self.iface.SetPopconPariticipation(False)
        with open(popcon_p) as f:
            popcon = f.read()
            self.assertNotIn('PARTICIPATE="yes"', popcon)
            self.assertIn('PARTICIPATE="no"', popcon)

    def test_updates_automation(self):
        states = [UPDATE_INST_SEC, UPDATE_DOWNLOAD, UPDATE_NOTIFY]
        # security
        self.iface.SetUpdateAutomationLevel(states[0])
        cfg = os.path.join(os.path.dirname(__file__),
                           "aptroot", "etc", "apt", "apt.conf.d",
                           "10periodic")
        with open(cfg) as f:
            config = f.read()
            self.assertIn('APT::Periodic::Unattended-Upgrade "1";', config)
        # download
        self.iface.SetUpdateAutomationLevel(states[1])
        with open(cfg) as f:
            config = f.read()
            self.assertIn('APT::Periodic::Unattended-Upgrade "0";', config)
            self.assertIn('APT::Periodic::Download-Upgradeable-Packages "1";', config)
        # notify
        self.iface.SetUpdateAutomationLevel(states[2])
        with open(cfg) as f:
            config = f.read()
            self.assertIn('APT::Periodic::Unattended-Upgrade "0";', config)
            self.assertIn('APT::Periodic::Download-Upgradeable-Packages "0";', config)

    def test_updates_interval(self):
        # interval
        self.iface.SetUpdateInterval(0)
        cfg = os.path.join(os.path.dirname(__file__),
                           "aptroot", "etc", "apt", "apt.conf.d",
                           "10periodic")
        with open(cfg) as f:
            config = f.read()
            self.assertTrue(
                'APT::Periodic::Update-Package-Lists' not in config or
                'APT::Periodic::Update-Package-Lists "0";' in config)
        self.iface.SetUpdateInterval(1)
        with open(cfg) as f:
            config = f.read()
            self.assertIn('APT::Periodic::Update-Package-Lists "1";', config)
        self.iface.SetUpdateInterval(0)
        with open(cfg) as f:
            config = f.read()
            self.assertIn('APT::Periodic::Update-Package-Lists "0";', config)

    def test_add_remove_source_by_line(self):
        # add invalid
        res = self.iface.AddSourceFromLine("xxx")
        self.assertFalse(res)
        # add real
        s = "deb https://ppa.launchpadcontent.net/ foo bar"
        self.iface.AddSourceFromLine(s)
        self.assertIn(s, self.sourceslist)
        self.assertIn(s.replace("deb", "# deb-src"), self.sourceslist)
        # remove again
        self.iface.RemoveSource(s)
        self.iface.RemoveSource(s.replace("deb", "deb-src"))
        self.assertNotIn(s, self.sourceslist)
        self.assertNotIn(s.replace("deb", "# deb-src"), self.sourceslist)
        self._assert_eventually("sources_list_count", 4)

    def test_add_gpg_key(self):
        # clean
        gpg_glob = os.path.join(os.path.dirname(__file__),
                              "aptroot", "etc", "apt", "*.gpg")
        trusted_gpg = os.path.join(os.path.dirname(__file__),
                                   "aptroot", "etc", "apt", "trusted.gpg")
        testkey = os.path.join(os.path.dirname(__file__), 
                               "data", "testkey.gpg")
        for f in glob.glob(gpg_glob):
            os.remove(f)
        self.assertTrue(len(glob.glob(gpg_glob)) == 0)
        # add key from file
        res = self.iface.AddKey(os.path.join(os.path.dirname(__file__),
                                             "data", "testkey.gpg"))
        self.assertTrue(res)
        self.assertEqual(len(glob.glob(gpg_glob)), 1)
        self.assertNotEqual(os.path.getsize(trusted_gpg), 0)
        # remove the key 
        res = self.iface.RemoveKey("D732CA59")
        self.assertTrue(res)
        if os.path.exists(trusted_gpg):
            self.assertEqual(os.path.getsize(trusted_gpg), 0)
        # add from data
        with open(testkey) as keyfile:
            data = keyfile.read()
            res = self.iface.AddKeyFromData(data)
            self.assertTrue(res)
            self.assertEqual(len(glob.glob(gpg_glob)), 1)
            self.assertNotEqual(os.path.getsize(trusted_gpg), 0)
        # remove the key 
        res = self.iface.RemoveKey("D732CA59")
        self.assertTrue(res)
        if os.path.exists(trusted_gpg):        
            self.assertEqual(os.path.getsize(trusted_gpg), 0)
        # test nonsense
        res = self.iface.AddKeyFromData("nonsens")
        self.assertFalse(res)
        # test apt-key update
        res = self.iface.UpdateKeys()
        self.assertTrue(res)
        

if __name__ == "__main__":
    if "-d" in sys.argv:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)
    unittest.main()



# Copyright (C) 2009 Canonical
#
# Authors:
#  Michael Vogt
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from __future__ import print_function

import aptsources.distro
from datetime import datetime
import distro_info
from functools import wraps
import gi
gi.require_version("Gtk", "3.0")
gi.require_version("Handy", "1")
from gi.repository import Gio, Gtk, GLib, Handy
import json
import os
import subprocess

import logging
LOG=logging.getLogger(__name__)

import time

UA_STATUS_JSON = "/var/lib/ubuntu-advantage/status.json"

def setup_ui(self, path, domain):
    # setup ui
    self.builder = Gtk.Builder()
    self.builder.set_translation_domain(domain)
    self.builder.add_from_file(path)
    self.builder.connect_signals(self)
    for o in self.builder.get_objects():
        if issubclass(type(o), Gtk.Buildable):
            name = Gtk.Buildable.get_name(o)
            setattr(self, name, o)
        else:
            logging.debug("can not get name for object '%s'" % o)

def has_gnome_online_accounts():
    try:
        d = Gio.DesktopAppInfo.new('gnome-online-accounts-panel.desktop')
        return d != None
    except Exception:
        return False

def is_current_distro_lts():
    distro = aptsources.distro.get_distro()
    di = distro_info.UbuntuDistroInfo()
    return di.is_lts(distro.codename)

def is_current_distro_supported():
    distro = aptsources.distro.get_distro()
    di = distro_info.UbuntuDistroInfo()
    return distro.codename in di.supported(datetime.now().date())

def current_distro():
    distro = aptsources.distro.get_distro()
    di = distro_info.UbuntuDistroInfo()
    releases = di.get_all(result="object")
    for release in releases:
        if release.series == distro.codename:
            return release


def get_ua_status():
    """Return a dict of all UA status information or empty dict on error."""
    # status.json will exist on any attached system. It will also be created
    # by the systemd timer ua-timer which will update UA_STATUS_JSON every 12
    # hours to reflect current status of UA subscription services.
    # Invoking `ua status` with subp will result in a network call to
    # contracts.canonical.com which could raise Timeouts on network limited
    # machines. So, prefer the status.json file when possible.
    status_json = ""
    if os.path.exists(UA_STATUS_JSON):
        with open(UA_STATUS_JSON) as stream:
            status_json = stream.read()
    else:
        try:
            # Success writes UA_STATUS_JSON
            result = subprocess.run(
                ['ua', 'status', '--format=json'], stdout=subprocess.PIPE
            )
        except Exception as e:
            print("Failed to run `ua status`:\n%s" % e)
            return {}
        if result.returncode != 0:
            print(
                "Ubuntu Advantage client returned code %d" % result.returncode
            )
            return {}
        status_json = result.stdout
    if not status_json:
        print(
            "Warning: no Ubuntu Advantage status found."
            " Is ubuntu-advantage-tools installed?"
        )
        return {}
    try:
        status = json.loads(status_json)
    except json.JSONDecodeError as e:
        print("Failed to parse ubuntu advantage client JSON:\n%s" % e)
        return {}
    if status.get("_schema_version", "0.1") != "0.1":
        print(
            "UA status schema version change: %s" % status["_schema_version"]
        )
    return status


def get_ua_service_status(service_name='esm-infra', status=None):
    """Get service availability and status for a specific UA service.

    Return a tuple (available, service_status).
      :boolean available: set True when either:
        - attached contract is entitled to the service
        - unattached machine reports service "availability" as "yes"
      :str service_status: will be one of the following:
        - "disabled" when the service is available and applicable but not
          active
        - "enabled" when the service is available and active
        - "n/a" when the service is not applicable to the environment or not
          entitled for the attached contract
    """
    if not status:
        status = get_ua_status()
    # Assume unattached on empty status dict
    available = False
    service_status = "n/a"
    for service in status.get("services", []):
        if service.get("name") != service_name:
            continue
        if "available" in service:
            available = bool("yes" == service["available"])
        if "status" in service:
            service_status = service["status"]  # enabled, disabled or n/a
    return (available, service_status)


def retry(exceptions, tries=10, delay=0.1, backoff=2):
    """
    Retry calling the decorated function using an exponential backoff.

    Args:
        exceptions: The exception to check. may be a tuple of
            exceptions to check.
        tries: Number of times to try (not retry) before giving up.
        delay: Initial delay between retries in seconds.
        backoff: Backoff multiplier (e.g. value of 2 will double the delay
            each retry).
    """
    def deco_retry(f):

        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except exceptions as e:
                    msg = '{}, Retrying in {} seconds...'.format(e, mdelay)
                    logging.warning(msg)
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)

        return f_retry  # true decorator

    return deco_retry

def is_dark_theme(widget):
    env_gtk_theme = GLib.getenv("GTK_THEME")
    if env_gtk_theme != None:
        return GLib.str_has_suffix(env_gtk_theme, "dark")
    return Handy.StyleManager.get_default().get_dark()



#!/usr/bin/python3

import threading
try:
    from queue import Empty, Queue
except ImportError:
    from Queue import Empty, Queue
import time
import re
import subprocess
import os
import aptsources.distro
import aptsources.sourceslist
try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen
import random

class MirrorTest(threading.Thread):
    """Determines the best mirrors by perfoming ping and download test."""
    class PingWorker(threading.Thread):
        """Use the command line command ping to determine the server's
           response time. Using multiple threads allows to run several
           test simultaneously."""
        def __init__(self, jobs, results, id, parent, borders=(0,1), mod=(0,0)):
            self.borders = borders
            self.mod = mod
            self.parent = parent
            self.id = id
            self.jobs = jobs
            self.results = results
            self.match_result = re.compile(r"^rtt .* = [\.\d]+/([\.\d]+)/.*")
            threading.Thread.__init__(self)
        def run(self):
            result = None
            while not self.jobs.empty() and self.parent.running.isSet():
                try:
                    mirror = self.jobs.get(False)
                    host = mirror.hostname

                    self.parent.report_action("Pinging %s..." % host)
                    commando = subprocess.Popen(["ping", "-q", "-c 2", "-W 1", "-i 0.5", host],
                                                stdout=subprocess.PIPE,
                                                stderr=subprocess.STDOUT,
                                                universal_newlines=True).stdout
                    while True:
                        line = commando.readline()
                        if not line:
                            break
                        result = re.findall(self.match_result, line)
                except Empty:
                    return
                except:
                    self.parent.report_action("Skipping %s" % host)
                # report and count the mirror (regardless of its success)
                MirrorTest.completed_lock.acquire()
                MirrorTest.completed += 1
                self.parent.report_progress(MirrorTest.completed,
                                            MirrorTest.todo,
                                            self.borders,
                                            self.mod)
                if result:
                    self.results.append([float(result[0]), host, mirror])
                MirrorTest.completed_lock.release()

    def __init__(self, mirrors, test_file, event, running=None):
        threading.Thread.__init__(self)
        self.action = ''
        self.progress = (0, 0, 0.0) # cur, max, %
        self.event = event
        self.best = None
        self.test_file = test_file
        self.threads = []
        MirrorTest.completed = 0
        MirrorTest.completed_lock = threading.Lock()
        MirrorTest.todo = len(mirrors)
        self.mirrors = mirrors
        if not running:
            self.running = threading.Event()
        else:
            self.running = running

    def report_action(self, text):
        self.action = text
        if self.event:
            self.event.set()

    def report_progress(self, current, max, borders=(0,100), mod=(0,0)):
        """Subclasses should override this method to receive
           progress status updates"""
        self.progress = (current + mod[0], 
                         max + mod[1],
                         borders[0] + (borders[1] - borders[0]) / max * current)
        if self.event:
            self.event.set()

    def run_full_test(self):
        # Determinate the 5 top ping servers
        results_ping = self.run_ping_test(max=5, borders=(0, 0.5), mod=(0,7))
        # Add two random mirrors to the download test
        size = len(self.mirrors)
        if size > 2:
            results_ping.append([0, 0, self.mirrors[random.randint(1, size-1)]])
            results_ping.append([0, 0, self.mirrors[random.randint(1, size-1)]])
        results = self.run_download_test([r[2] for r in results_ping],
                                         borders=(0.5, 1),
                                         mod=(MirrorTest.todo,
                                              MirrorTest.todo))
        for (t, h) in results:
            print("mirror: %s - time: %s" % (h.hostname, t))
        if not results:
            return None
        else:
            winner = results[0][1].hostname
            print("and the winner is: %s" % winner)
            return winner

    def run_ping_test(self, mirrors=None, max=None, borders=(0,1), mod=(0,0)):
        """Performs ping tests of the given mirrors and returns the
           best results (specified by max).
           Mod and borders could be used to tweak the reported result if
           the download test is only a part of a whole series of tests."""
        if mirrors == None:
            mirrors = self.mirrors
        jobs = Queue()
        for m in mirrors:
            jobs.put(m)
        results = []
        #FIXME: Optimze the number of ping working threads LP#90379
        for i in range(25):
            t = MirrorTest.PingWorker(jobs, results, i, self, borders, mod)
            self.threads.append(t)
            t.start()

        for t in self.threads:
            t.join()

        results.sort()
        return results[0:max]

    def run_download_test(self, mirrors=None, max=None, borders=(0,1), 
                          mod=(0,0)):
        """Performs download tests of the given mirrors and returns the
           best results (specified by max).
           Mod and borders could be used to tweak the reported result if
           the download test is only a part of a whole series of tests."""
        def test_download_speed(mirror):
            url = "%s/%s" % (mirror.get_repo_urls()[0],
                             self.test_file)
            self.report_action("Downloading %s..." % url)
            start = time.time()
            try:
                urlopen(url, timeout=2).read(102400)
                return time.time() - start
            except:
                return 0
        if mirrors == None:
            mirrors = self.mirrors
        results = []

        for m in mirrors:
            if not self.running.isSet():
                break
            download_time = test_download_speed(m)
            if download_time > 0:
                results.append([download_time, m])
            self.report_progress(mirrors.index(m) + 1, len(mirrors), (0.50,1), mod)
        results.sort()
        return results[0:max]

    def run(self):
        """Complete test exercise, set self.best when done"""
        self.best = self.run_full_test()
        self.running.clear()

if __name__ == "__main__":
    distro = aptsources.distro.get_distro()
    distro.get_sources(aptsources.sourceslist.SourcesList())
    pipe = os.popen("dpkg --print-architecture")
    arch = pipe.read().strip()
    running = threading.Event()
    running.set()
    test_file = "dists/%s/%s/binary-%s/Packages.gz" % \
                (distro.source_template.name,
                 distro.source_template.components[0].name,
                 arch)
    app = MirrorTest(list(distro.source_template.mirror_set.values()),
                     test_file, threading.Event(), running)
    app.run_full_test()



#!/usr/bin/python3

import os
import re
import shutil
import subprocess
import tempfile
import unittest

from aptsources.distro import get_distro
from copy import copy
from pathlib import Path
from types import SimpleNamespace
from urllib.request import urlopen
from urllib.error import (URLError, HTTPError)

from softwareproperties.cloudarchive import RELEASE_MAP
from softwareproperties.extendedsourceslist import SourceEntry


CODENAME = get_distro().codename

ADD_APT_REPOSITORY = shutil.which('add-apt-repository', path=f'.:{os.environ.get("PATH")}')

FAKE_URI = 'http://fake.archive.test.com/ubuntu'
SOURCESLIST_TEXT = f'deb {FAKE_URI} {CODENAME} main'

# types
TYPE_BINARY = 'deb'
TYPE_SOURCE = 'deb-src'
# pockets
RELEASE = 'release'
PROPOSED = 'proposed'
UPDATES = 'updates'
SECURITY = 'security'
BACKPORTS = 'backports'
# components
MAIN = frozenset(['main'])
UNIVERSE = frozenset(['universe'])
RESTRICTED = frozenset(['restricted'])
MULTIVERSE = frozenset(['multiverse'])

ARCHIVE_URI = 'http://fake.mirror.private.com/ubuntu/'
ARCHIVE_FILENAME_URI = re.sub(r'[:/.]+', '_', ARCHIVE_URI)
ARCHIVE_FILENAME = f'archive_uri-{ARCHIVE_FILENAME_URI}-{CODENAME}.list'

PPA_NAME = 'ubuntu-support-team/software-properties-autopkgtest'
PPA_URI = f'https://ppa.launchpadcontent.net/{PPA_NAME}/ubuntu/'
PPA_FILENAME = f'ubuntu-support-team-ubuntu-software-properties-autopkgtest-{CODENAME}.list'
PPA_TRUSTED_FILENAME = 'ubuntu-support-team-ubuntu-software-properties-autopkgtest.gpg'
PPA_TRUSTED_FINGERPRINT = 'A17A D76F CBB7 A7D5 73C4  DF43 1E03 2FCE 2F88 6048'

ADD_APT_REPOSITORY_LONG_PARAMS = SimpleNamespace(
    remove='--remove',
    enable_source='--enable-source',
    component='--component',
    pocket='--pocket',
    yes='--yes',
    no_update='--no-update',
    ppa='--ppa',
    cloud='--cloud',
    uri='--uri',
    sourceslist='--sourceslist',
)

ADD_APT_REPOSITORY_SHORT_PARAMS = SimpleNamespace(
    remove='-r',
    enable_source='-s',
    component='-c',
    pocket='-p',
    yes='-y',
    no_update='-n',
    ppa='-P',
    cloud='-C',
    uri='-U',
    sourceslist='-S',
)


def has_network():
    try:
        with urlopen("https://launchpad.net/"):
            pass
    except (URLError, HTTPError):
        return False
    return True

def line(uri=None, binary=True, pocket=None, components=None):
    deb = TYPE_BINARY if binary else TYPE_SOURCE
    uri = uri or FAKE_URI
    pocket = f'-{pocket}' if pocket else ''
    components = ' '.join(components or MAIN)
    return f'{deb} {uri} {CODENAME}{pocket} {components}'


@unittest.skipIf(ADD_APT_REPOSITORY is None, 'Could not find add-apt-repository script')
class TestAddAptRepository(unittest.TestCase):
    params = ADD_APT_REPOSITORY_LONG_PARAMS

    @classmethod
    def setUpClass(cls):
        # test apt.conf file
        d = tempfile.TemporaryDirectory()
        cls.addClassCleanup(d.cleanup)
        apt_conf = Path(d.name) / 'apt.conf'

        # set APT_CONFIG to the test apt.conf
        test_env = copy(os.environ)
        test_env['APT_CONFIG'] = str(apt_conf)

        # dir for test apt-get script
        d = tempfile.TemporaryDirectory()
        cls.addClassCleanup(d.cleanup)
        apt_get_dir = Path(d.name)

        # create a test apt-get script and prepend it to the PATH
        apt_get_output = apt_get_dir / 'test-apt-get-output'
        apt_get_script = apt_get_dir / 'apt-get'
        apt_get_script.touch(mode=0o755)
        apt_get_script.write_text(f'''#!/bin/bash\necho "$*" >> {apt_get_output}\n''')
        test_env['PATH'] = f'{apt_get_dir}:{test_env["PATH"]}'

        cls.apt_conf = apt_conf
        cls.apt_get_output = apt_get_output
        cls.test_env = test_env

    def setUp(self):
        d = tempfile.TemporaryDirectory()
        self.addCleanup(d.cleanup)
        dir_etc = Path(d.name)
        self.apt_conf.write_text(f'Dir::Etc "{dir_etc}";')

        dir_etc.joinpath('apt.conf.d').mkdir(exist_ok=True)

        self.sourceslist = dir_etc / 'sources.list'
        self.sourceslist.touch()
        self.sourceslist.write_text(SOURCESLIST_TEXT)

        self.sourceslistd = dir_etc / 'sources.list.d'
        self.sourceslistd.mkdir(exist_ok=True)

        self.trustedgpgd = dir_etc / 'trusted.gpg.d'
        self.trustedgpgd.mkdir(exist_ok=True)

        self._sourcesfile = None
        self._trustedfile = None

        self.test_env.pop('LC_ALL', None)

        self.testing_ppa = False

        self.testing_uca = False
        self.test_env.pop('CA_ALLOW_CODENAME', None)

        self.noupdate = False
        self.pocket = None
        self.components = None
        self.enable_source = 0

    @property
    def sourcesfile(self):
        return self._sourcesfile

    @property
    def trustedfile(self):
        return self._trustedfile

    @sourcesfile.setter
    def sourcesfile(self, filename):
        self._sourcesfile = self._file_setter(self.sourceslistd, filename)

    @trustedfile.setter
    def trustedfile(self, filename):
        self._trustedfile = self._file_setter(self.trustedgpgd, filename)

    def _file_setter(self, d, f):
        if not f:
            return None
        path = d / f
        self.addCleanup(path.unlink, missing_ok=True)
        return path

    @property
    def parameter_enable_source(self):
        return [self.params.enable_source] * self.enable_source

    @property
    def parameter_pocket(self):
        return [self.params.pocket, self.pocket] if self.pocket else []

    @property
    def parameter_components(self):
        params = []
        for c in self.components or []:
            params.append(self.params.component)
            params.append(c)
        return params

    @property
    def parameter_all(self):
        return self.parameter_enable_source + self.parameter_pocket + self.parameter_components

    def line(self, uri=None, binary=True, pocket=None, components=None):
        return line(uri=uri or ARCHIVE_URI,
                    binary=binary,
                    pocket=pocket or self.pocket,
                    components=components or self.components)

    @unittest.skipUnless(os.geteuid() == 0, 'requires root')
    def call_add_apt_repository(self, args):
        self.apt_get_output.unlink(missing_ok=True)

        n = [self.params.no_update] if self.noupdate else []
        result = subprocess.run([ADD_APT_REPOSITORY] + ['-y'] + n + args,
                                stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                env=self.test_env)

        if result.returncode != 0:
            self.fail(f'add-apt-repository failed: {result.stdout}')

    def check_sourceentry(self, expected_line, content_lines):
        expected_entry = SourceEntry(expected_line)

        for (index, line) in enumerate(content_lines):
            entry = SourceEntry(line)
            if entry == expected_entry:
                return content_lines[:index] + content_lines[index+1:]

        content = '\n'.join(content_lines)
        self.fail(f"Missing line: '{expected_line}' in: '{content}'")

    def check_expected_lines(self, expected_lines, lines):
        for expected_line in expected_lines:
            lines = self.check_sourceentry(expected_line, lines)

        # ignore invalid or disabled lines
        lines = list(filter(lambda l: not SourceEntry(l).invalid, lines))
        lines = list(filter(lambda l: not SourceEntry(l).disabled, lines))
        if lines:
            unexpected_lines = '\n'.join(lines)
            self.fail(f"Unexpected lines:\n{unexpected_lines}")

    def apt_get_output_lines(self):
        try:
            return self.apt_get_output.read_text().splitlines()
        except FileNotFoundError:
            return []

    def check_apt_get_update(self):
        if self.noupdate:
            self.assertNotIn('update', self.apt_get_output_lines())
        else:
            self.assertIn('update', self.apt_get_output_lines())

    def check_uca_keyring(self):
        if self.testing_uca:
            self.assertIn('install -y ubuntu-cloud-keyring', self.apt_get_output_lines())

    def check_ppa_gpgkey(self):
        if not self.testing_ppa:
            return

        self.assertIsNotNone(self.trustedfile)

        with tempfile.TemporaryDirectory() as homedir:
            Path(homedir).chmod(0o700) # so gpg doesn't complain to stderr
            cmd = f'gpg -q --homedir {homedir} --no-default-keyring --keyring {self.trustedfile} --fingerprint'
            result = subprocess.run(cmd.split(), stdout=subprocess.PIPE, encoding='UTF-8')
            self.assertIn(PPA_TRUSTED_FINGERPRINT, result.stdout)

    def check_file(self, testfile, exists=True):
        if not testfile:
            return
        msg = f'{"missing" if exists else "unexpected"} {testfile}'
        self.assertEqual(testfile.exists(), exists, msg)

    def run_sourceslistd_add_remove_test(self, args, expected_lines=[]):
        self.check_file(self.sourcesfile, exists=False)
        self.check_file(self.trustedfile, exists=False)

        self.call_add_apt_repository(args)

        self.check_file(self.sourcesfile)
        self.check_file(self.trustedfile)

        self.check_expected_lines(expected_lines, self.sourcesfile.read_text().splitlines())

        self.check_apt_get_update()
        self.check_ppa_gpgkey()
        self.check_uca_keyring()

        self.call_add_apt_repository([self.params.remove] + args)

        self.check_file(self.sourcesfile, exists=False)

        # trusted.gpg.d file isn't removed by add-apt-repository
        self.check_file(self.trustedfile)

    def run_sourceslistd_test(self, args, expected_line):
        source_entry = SourceEntry(expected_line)._replace(type=TYPE_SOURCE)
        source_entry.set_enabled(self.enable_source > 0)

        expected_lines = [expected_line, str(source_entry)]
        test_args = self.parameter_all
        test_args += args
        self.run_sourceslistd_add_remove_test(test_args, expected_lines)

    def test_uri(self):
        self.sourcesfile = ARCHIVE_FILENAME
        args = [self.params.uri, ARCHIVE_URI]
        self.run_sourceslistd_test(args, self.line())

    def test_uri_noupdate(self):
        self.noupdate = True
        self.test_uri()

    def test_sourceslist(self):
        self.sourcesfile = ARCHIVE_FILENAME
        line = self.line()
        args = [self.params.sourceslist, line]
        self.run_sourceslistd_test(args, line)

    def test_sourceslist_noupdate(self):
        self.noupdate = True
        self.test_sourceslist()

    @unittest.skipUnless(has_network(), 'requires network')
    def test_ppa(self, /, with_param=True, prefix=False):
        self.testing_ppa = True

        self.sourcesfile = PPA_FILENAME
        self.trustedfile = PPA_TRUSTED_FILENAME
        args = []
        if with_param:
            args += [self.params.ppa]
        if prefix:
            args += [f'ppa:{PPA_NAME}']
        else:
            args += [PPA_NAME]
        self.run_sourceslistd_test(args, self.line(uri=PPA_URI))

    def test_ppa_noupdate(self):
        self.noupdate = True
        self.test_ppa()

    def test_ppa_with_prefix(self):
        self.test_ppa(prefix=True)

    def test_ppa_without_param_with_prefix(self):
        self.test_ppa(with_param=False, prefix=True)

    def _test_uca(self, uca_release, /, with_param, prefix):
        self.testing_uca = True
        if CODENAME != RELEASE_MAP.get(uca_release):
            self.test_env['CA_ALLOW_CODENAME'] = CODENAME

        self.sourcesfile = f'cloudarchive-{uca_release}.list'
        line = f'deb http://ubuntu-cloud.archive.canonical.com/ubuntu {CODENAME}-updates/{uca_release} main'
        args = []
        if with_param:
            args += [self.params.cloud]
        if prefix:
            args += [f'{prefix}:{uca_release}']
        else:
            args += [uca_release]
        self.run_sourceslistd_test(args, line)

    def test_uca(self, /, with_param=True, prefix=None):
        uca_releases = list(RELEASE_MAP.keys())
        self.assertGreater(len(uca_releases), 0)
        for uca_release in uca_releases:
            with self.subTest(uca_release=uca_release):
                self.setUp()
                self._test_uca(uca_release, with_param=with_param, prefix=prefix)
                self.tearDown()

    def test_uca_noupdate(self):
        self.noupdate = True
        self.test_uca()

    def test_uca_with_prefix_uca(self):
        self.test_uca(prefix='uca')

    def test_uca_with_prefix_uca_without_param(self):
        self.test_uca(with_param=False, prefix='uca')

    def test_uca_with_prefix_cloud_archive(self):
        self.test_uca(prefix='cloud-archive')

    def test_uca_with_prefix_cloud_archive_without_param(self):
        self.test_uca(with_param=False, prefix='cloud-archive')

    class TestLinesGroup(object):
        def __init__(self, /, test_pocket=None, test_components=frozenset(), repo_n=1):
            self.test_pocket = test_pocket
            self.test_components = test_components
            self.repo_n = repo_n

            self.lines = []
            self.expected_lines = []
            self.after_remove = []

        def next_repo(self):
            self.repo_n += 1

        def entry(self, /, binary=True, pocket=None, components=MAIN):
            return SourceEntry(self.line(binary=binary, pocket=pocket, components=components))

        def line(self, /, binary=True, pocket=None, components=MAIN):
            n = self.repo_n
            return line(uri=f'{FAKE_URI}/source{n}', binary=binary, pocket=pocket, components=components)

    def check_lines_group(self, args, group):
        self.sourceslist.write_text('\n'.join(group.lines))

        self.call_add_apt_repository(args)
        self.check_expected_lines(group.expected_lines,
                                  self.sourceslist.read_text().splitlines())


        self.call_add_apt_repository([self.params.remove] + args)
        self.check_expected_lines(group.after_remove,
                                  self.sourceslist.read_text().splitlines())

    def add_lines_comp(self, group, components, first_line=True, other_lines=frozenset()):
        group.lines.append(group.line(components=components))

        expected = copy(components)
        if first_line:
            # all new components will be added to (only) the first line
            expected |= group.test_components
            expected -= other_lines
        group.expected_lines.append(group.line(components=expected))

        after_remove = components - group.test_components
        if after_remove:
            group.after_remove.append(group.line(components=after_remove))

    def test_component(self):
        # with no self.components, we test the no-prefix case using MAIN
        test_components = copy(self.components or MAIN)
        args = self.parameter_components or list(test_components)

        group = self.TestLinesGroup(test_components=test_components)

        # source1, with two lines, one component on each
        self.add_lines_comp(group, MAIN, other_lines=UNIVERSE)
        self.add_lines_comp(group, UNIVERSE, first_line=False)

        # source2, one line with two components
        group.next_repo()
        self.add_lines_comp(group, MAIN | UNIVERSE)

        # source3, same as source2 but different components
        group.next_repo()
        self.add_lines_comp(group, RESTRICTED | MULTIVERSE)

        # source4, one line, one component
        group.next_repo()
        self.add_lines_comp(group, RESTRICTED)

        self.check_lines_group(args, group)

    def test_component_main(self):
        self.components = MAIN
        self.test_component()

    def test_component_main_noupdate(self):
        self.noupdate = True
        self.test_component_main()

    def test_component_universe(self):
        self.components = UNIVERSE
        self.test_component()

    def test_component_universe_noupdate(self):
        self.noupdate = True
        self.test_component_universe()

    def test_component_restricted(self):
        self.components = RESTRICTED
        self.test_component()

    def test_component_multiverse(self):
        self.components = MULTIVERSE
        self.test_component()

    def test_component_main_universe(self):
        self.components = MAIN | UNIVERSE
        self.test_component()

    def test_component_universe_multiverse(self):
        self.components = UNIVERSE | MULTIVERSE
        self.test_component()

    def test_component_all(self):
        self.components = MAIN | UNIVERSE | RESTRICTED | MULTIVERSE
        self.test_component()

    def test_component_all_except_main(self):
        self.components = UNIVERSE | RESTRICTED | MULTIVERSE
        self.test_component()

    def add_lines_pocket(self, group, pocket, components=MAIN, other_pockets=[], other_comps=frozenset(), first_line=True):
        line = group.line(pocket=pocket, components=components)
        group.lines.append(line)
        group.expected_lines.append(line)
        if (pocket or RELEASE) != group.test_pocket:
            group.after_remove.append(line)

    def _test_pocket(self):
        self.assertIsNotNone(self.pocket)

        args = self.parameter_pocket

        group = self.TestLinesGroup(test_pocket=self.pocket)

        # source1, with two lines, one pocket, one component on each
        self.add_lines_pocket(group, None, MAIN)
        self.add_lines_pocket(group, None, UNIVERSE)
        if self.pocket not in [RELEASE]:
            group.expected_lines.append(group.line(pocket=self.pocket, components=MAIN | UNIVERSE))

        # source2, with two lines, two pockets, varying components on each
        group.next_repo()
        self.add_lines_pocket(group, None, MAIN | UNIVERSE)
        self.add_lines_pocket(group, UPDATES, MAIN)
        if self.pocket not in [RELEASE, UPDATES]:
            group.expected_lines.append(group.line(pocket=self.pocket, components=MAIN | UNIVERSE))

        # source3, with three lines, two pockets, varying components on each
        group.next_repo()
        self.add_lines_pocket(group, None, MAIN)
        self.add_lines_pocket(group, None, UNIVERSE)
        self.add_lines_pocket(group, UPDATES, MAIN)
        if self.pocket not in [RELEASE, UPDATES]:
            group.expected_lines.append(group.line(pocket=self.pocket, components=MAIN | UNIVERSE))

        # source4, no release pocket
        group.next_repo()
        self.add_lines_pocket(group, UPDATES, MAIN)
        if self.pocket not in [UPDATES]:
            group.expected_lines.append(group.line(pocket=self.pocket, components=MAIN | RESTRICTED))

        # source5, all pockets exist
        group.next_repo()
        self.add_lines_pocket(group, None, MAIN)
        self.add_lines_pocket(group, UPDATES, MAIN)
        self.add_lines_pocket(group, PROPOSED, MAIN)
        self.add_lines_pocket(group, SECURITY, MAIN)

        # source6, release lines specified with and without pocket
        group.next_repo()
        self.add_lines_pocket(group, None, MAIN)
        self.add_lines_pocket(group, RELEASE, MULTIVERSE)
        self.add_lines_pocket(group, UPDATES, MAIN)
        if self.pocket not in [RELEASE, UPDATES]:
            group.expected_lines.append(group.line(pocket=self.pocket, components=MAIN | MULTIVERSE))

        self.check_lines_group(args, group)

    def test_pocket_release(self):
        self.pocket = RELEASE
        self._test_pocket()

    def test_pocket_release_noupdate(self):
        self.noupdate = True
        self.test_pocket_release()

    def test_pocket_updates(self):
        self.pocket = UPDATES
        self._test_pocket()

    def test_pocket_updates_noupdate(self):
        self.noupdate = True
        self.test_pocket_updates()

    def test_pocket_proposed(self):
        self.pocket = PROPOSED
        self._test_pocket()

    def test_pocket_security(self):
        self.pocket = SECURITY
        self._test_pocket()

    def add_lines_es(self, group, pocket, components, /, binary=True, disabled=False, alone=False, plus_comps=[], minus_comps=False):
        binary_entry = group.entry(pocket=pocket, components=components)._replace(disabled=disabled)
        enabled_source_entry = group.entry(binary=False, pocket=pocket, components=components)
        disabled_source_entry = enabled_source_entry._replace(disabled=True)

        if binary:
            binary_line = str(binary_entry)
            group.lines.append(binary_line)
            group.expected_lines.append(binary_line)
            group.after_remove.append(binary_line)
            # with binary, 'alone' indicates no matching source line exists,
            # so we need to create one if enable_source > 1
            # if minus_comps is True, our comps will be added to another line
            if alone and not disabled and not minus_comps and self.enable_source > 1:
                group.expected_lines.append(str(enabled_source_entry))
                group.after_remove.append(str(disabled_source_entry))
        else:
            enabled_line = str(enabled_source_entry)
            disabled_line = str(disabled_source_entry)
            if not disabled:
                group.lines.append(enabled_line)
            else:
                group.lines.append(disabled_line)
            if plus_comps and self.enable_source > 1:
                # include comps from other binary-only lines
                comps = set(enabled_source_entry.comps) | set(plus_comps)
                enabled_line = str(enabled_source_entry._replace(comps=comps))
                disabled_line = str(disabled_source_entry._replace(comps=comps))
            if disabled and alone:
                # with source, 'alone' indicates no matching (enabled) binary line exists,
                # so the source line will *not* be enabled
                group.expected_lines.append(disabled_line)
            else:
                # if already enabled, or if not 'alone', we expect it to be enabled
                group.expected_lines.append(enabled_line)
            group.after_remove.append(disabled_line)

    def add_lines_es_binary(self, group, pocket, components, /, **kwargs):
        self.add_lines_es(group, pocket, components, binary=True, **kwargs)

    def add_lines_es_source(self, group, pocket, components, /, **kwargs):
        self.add_lines_es(group, pocket, components, binary=False, **kwargs)

    def add_lines_es_both(self, group, pocket, components, /, source_disabled=False, **kwargs):
        self.add_lines_es_binary(group, pocket, components, **kwargs)
        if kwargs.get('disabled'):
            # if both lines are disabled, source line is effectively 'alone'
            # as there is no enabled matching binary line
            kwargs['alone'] = True
        if source_disabled:
            kwargs['disabled'] = True
        self.add_lines_es_source(group, pocket, components, **kwargs)

    def _test_enable_source(self):
        self.assertGreater(self.enable_source, 0)

        args = self.parameter_enable_source

        group = self.TestLinesGroup()

        # source1, simple test with just binary lines
        self.add_lines_es_binary(group, RELEASE, MAIN | UNIVERSE, alone=True)
        self.add_lines_es_binary(group, UPDATES, MAIN | UNIVERSE, alone=True)
        self.add_lines_es_binary(group, PROPOSED, MAIN | UNIVERSE, alone=True)
        self.add_lines_es_binary(group, SECURITY, MAIN | UNIVERSE, alone=True)

        # source2, all with matching disabled source lines
        group.next_repo()
        self.add_lines_es_both(group, None, MAIN | UNIVERSE, source_disabled=True)
        self.add_lines_es_both(group, None, RESTRICTED | MULTIVERSE, source_disabled=True)
        self.add_lines_es_both(group, UPDATES, MAIN | UNIVERSE | RESTRICTED | MULTIVERSE, source_disabled=True)
        self.add_lines_es_both(group, SECURITY, MAIN | UNIVERSE, source_disabled=True)

        # source3, mix of just binary lines plus some disabled and enabled source lines
        group.next_repo()
        self.add_lines_es_both(group, None, MAIN | UNIVERSE, source_disabled=True, plus_comps=RESTRICTED | MULTIVERSE)
        self.add_lines_es_binary(group, None, RESTRICTED | MULTIVERSE, alone=True, minus_comps=True)
        self.add_lines_es_both(group, UPDATES, MAIN | UNIVERSE | RESTRICTED | MULTIVERSE, source_disabled=True)
        self.add_lines_es_both(group, SECURITY, MAIN | UNIVERSE)

        # source4, disabled binary lines to verify disabled source lines aren't enabled
        group.next_repo()
        self.add_lines_es_both(group, None, MAIN | UNIVERSE, disabled=True)
        self.add_lines_es_both(group, UPDATES, MULTIVERSE, disabled=True)

        self.check_lines_group(args, group)

    def test_enable_source_1(self):
        self.enable_source = 1
        self._test_enable_source()

    def test_enable_source_1_noupdate(self):
        self.no_update = True
        self.test_enable_source_1()

    def test_enable_source_2(self):
        self.enable_source = 2
        self._test_enable_source()

    def test_enable_source_2_noupdate(self):
        self.no_update = True
        self.test_enable_source_2()

    @classmethod
    def add_locale_test(cls, testname, locale):
        newname = f"{testname}_locale_{re.sub(r'[-.]+', '_', locale)}"
        if hasattr(cls, newname):
            return
        def _test_locale(self):
            self.test_env['LC_ALL'] = locale
            getattr(self, testname)()
        setattr(cls, newname, _test_locale)

    @classmethod
    def add_combination_test(cls, testname, /, enable_source=0, pocket=None, components=None):
        newname = testname
        if enable_source > 0:
            newname += f'_enable_source_{enable_source}'
        if pocket:
            newname += f'_pocket_{pocket}'
        if components:
            newname += f'_components_{"_".join(components)}'
        if hasattr(cls, newname):
            return
        def _test_combination(self):
            self.enable_source = enable_source
            self.pocket = pocket
            self.components = components
            getattr(self, testname)()
        setattr(cls, newname, _test_combination)


class TestAddAptRepositoryShortParams(TestAddAptRepository):
    params = ADD_APT_REPOSITORY_SHORT_PARAMS


# Add various combination tests
for test in ['uri', 'sourceslist']:
    for locale in ['C', 'C.UTF-8']:
        TestAddAptRepository.add_locale_test(f'test_{test}', locale=locale)
    for enable_source in [0, 1, 2]:
        for pocket in [None, RELEASE, UPDATES, PROPOSED, SECURITY]:
            for components in [None, MAIN, UNIVERSE, RESTRICTED, MULTIVERSE, MAIN | UNIVERSE,
                               MAIN | UNIVERSE | RESTRICTED | MULTIVERSE]:
                TestAddAptRepository.add_combination_test(f'test_{test}',
                                                          enable_source=enable_source,
                                                          pocket=pocket,
                                                          components=components)


if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Qt 4 based frontend to software-properties
#
#  Copyright (c) 2007 Canonical Ltd.
#
#  Author: Jonathan Riddell <jriddell@ubuntu.com>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

from __future__ import absolute_import, print_function

from gettext import gettext as _
import os
import threading

from softwareproperties.MirrorTest import MirrorTest
from .I18nHelper import *

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5 import uic

from softwareproperties.CountryInformation import CountryInformation

class DialogMirror(QDialog):

    # Signals
  test_start = pyqtSignal()
  test_end = pyqtSignal('QString')
  report_progress = pyqtSignal(int, int, tuple, tuple)
  report_action = pyqtSignal('QString')

  def __init__(self, parent, datadir, distro, custom_mirrors):
    """
    Initialize the dialog that allows to choose a custom or official mirror
    """
    QDialog.__init__(self, parent)
    uic.loadUi("%s/designer/dialog_mirror.ui" % datadir, self)
    self.parent = parent

    self.custom_mirrors = custom_mirrors

    self.country_info = CountryInformation()

    self.button_choose = self.buttonBox.button(QDialogButtonBox.Ok)
    self.button_choose.setEnabled(False)
    self.button_cancel = self.buttonBox.button(QDialogButtonBox.Cancel)
    self.distro = distro

    self.treeview.setColumnCount(1)
    self.treeview.setHeaderLabels([_("Mirror")])

    translate_widget(self)
    # used to find the corresponding iter of a location
    map_loc = {}
    patriot = None
    """ no custom yet
    # at first add all custom mirrors and a separator
    if len(self.custom_mirrors) > 0:
        for mirror in self.custom_mirrors:

            model.append(None, [mirror, False, True, None])
            self.column_mirror.add_attribute(self.renderer_mirror,
                                             "editable",
                                             COLUMN_CUSTOM)
        model.append(None, [None, True, False, None])
    """
    self.mirror_map = {}
    # secondly add all official mirrors
    for hostname in self.distro.source_template.mirror_set.keys():
        mirror = self.distro.source_template.mirror_set[hostname]
        if mirror.location in map_loc:  # a mirror in a country
            QTreeWidgetItem(map_loc[mirror.location], [hostname])
            self.mirror_map[hostname] = mirror
        elif mirror.location != None:  # a country new to the list
            country = self.country_info.get_country_name(mirror.location)
            parent = QTreeWidgetItem([country])
            self.mirror_map[country] = None
            self.treeview.addTopLevelItem(parent)
            QTreeWidgetItem(parent, [hostname])
            self.mirror_map[hostname] = mirror
            if mirror.location == self.country_info.code and patriot == None:
                patriot = parent
            map_loc[mirror.location] = parent
        else:  # a mirror without country
            item = QTreeWidgetItem([hostname])
            self.treeview.addTopLevelItem(item)
    self.treeview.sortItems(0, Qt.AscendingOrder)
    # Scroll to the local mirror set
    if patriot != None:
        self.select_mirror(patriot.text(0))
    self.treeview.itemClicked.connect(self.on_treeview_mirrors_cursor_changed)
    self.button_find_server.clicked.connect(self.on_button_test_clicked)
    self.edit_buttons_frame.hide()  ##FIXME not yet implemented

  def on_treeview_mirrors_cursor_changed(self, item, column):
    """ Check if the currently selected row in the mirror list
        contains a mirror and or is editable """
    # Update the list of available protocolls
    hostname = self.treeview.currentItem().text(0)
    mirror = self.mirror_map[hostname]
    self.combobox.clear()
    if mirror != None:
        self.combobox.setEnabled(True)
        seen_protos = []
        self.protocol_paths = {}
        for repo in mirror.repositories:
            # Only add a repository for a protocoll once
            if repo.proto in seen_protos:
                continue
            seen_protos.append(repo.proto)
            self.protocol_paths[repo.get_info()[0]] = repo.get_info()[1]
            self.combobox.addItem(repo.get_info()[0])
        self.button_choose.setEnabled(True)
    else:
        self.button_choose.setEnabled(False)
    """
        # Allow to edit and remove custom mirrors
        self.button_remove.set_sensitive(model.get_value(iter, COLUMN_CUSTOM))
        self.button_edit.set_sensitive(model.get_value(iter, COLUMN_CUSTOM))
        self.button_choose.set_sensitive(self.is_valid_mirror(model.get_value(iter, COLUMN_URI)))
        self.combobox.set_sensitive(False)

    """

  def run(self):
    """ Run the chooser dialog and return the chosen mirror or None """
    res = self.exec_()

    # FIXME: we should also return the list of custom servers
    if res == QDialog.Accepted:
        hostname = self.treeview.currentItem().text(0)
        mirror = self.mirror_map[hostname]

        if mirror == None:
            # Return the URL of the selected custom mirror
            print("Error, unknown mirror")
            return None
            ##FIXME return model.get_value(iter, COLUMN_URI)
        else:
            # Return a URL created from the hostname and the selected
            # repository
            proto = self.combobox.currentText()

            directory = self.protocol_paths[proto]
            return "%s://%s/%s" % (proto, mirror.hostname, directory)
    else:
        return None

  def on_button_test_clicked(self):
    ''' Perform a test to find the best mirror and select it
        afterwards in the treeview '''
    class MirrorTestQt(MirrorTest):
        def __init__(self, mirrors, test_file, running, dialog, parent):
            MirrorTest.__init__(self, mirrors, test_file, None, running)
            self.dialog = dialog
            self.parent = parent

        def report_action(self, text):
            if self.running.isSet():
                self.parent.report_action.emit(text)

        def report_progress(self, current, max, borders=(0,1), mod=(0,0)):
            if self.running.isSet():
                self.parent.report_progress.emit(current, max, borders, mod)

        def run(self):
            self.parent.test_start.emit()
            rocker = self.run_full_test()
            self.parent.test_end.emit(rocker)

    self.dialog = QProgressDialog(_("Testing Mirrors"), _("Cancel"), 0, 100, self)
    self.dialog.setWindowTitle(_("Testing Mirrors"))
    self.dialog.setWindowModality(Qt.WindowModal)
    self.button_cancel_test = QPushButton(_("Cancel"), self.dialog)
    self.dialog.setCancelButton(self.button_cancel_test)
    self.button_cancel_test.clicked.connect(self.on_button_cancel_test_clicked);
    # the following signals are connected across threads
    self.test_start.connect(self.on_test_start, Qt.BlockingQueuedConnection)
    self.test_end.connect(self.on_test_end, Qt.BlockingQueuedConnection)
    self.report_progress.connect(self.on_report_progress, Qt.BlockingQueuedConnection)
    self.report_action.connect(self.on_report_action, Qt.BlockingQueuedConnection)

    self.running = threading.Event()
    self.running.set()
    pipe = os.popen("dpkg --print-architecture")
    arch = pipe.read().strip()
    test_file = "dists/%s/%s/binary-%s/Packages.gz" % \
                 (self.distro.source_template.name,
                  self.distro.source_template.components[0].name,
                  arch)
    test = MirrorTestQt(list(self.distro.source_template.mirror_set.values()),
                         test_file, self.running, self.dialog, self)
    test.start() # test starts in a separate thread

  def on_test_start(self):
      self.dialog.show()

  def on_test_end(self, mirror):
      self.dialog.hide()
      if not self.running.isSet():
          return # canceled by user
      if mirror != None:
          self.select_mirror(mirror)
      else:
          QMessageBox.warning(self.dialog, _("Testing Mirrors"),
                                  _("No suitable download server was found") + "\n" +
                                  _("Please check your Internet connection."))

  def on_report_action(self, text):
      self.dialog.setLabelText(str("<i>%s</i>" % text))

  def on_report_progress(self, current, max, borders=(0,1), mod=(0,0)):
      #self.dialog.setLabelText(_("Completed %s of %s tests") % \
      #                       (current + mod[0], max + mod[1]))
      frac = borders[0] + (borders[1] - borders[0]) / max * current
      self.dialog.setValue(frac*100)

  def on_button_cancel_test_clicked(self):
    ''' Abort the mirror performance test '''
    self.running.clear()
    self.dialog.show()
    self.dialog.setLabelText("<i>%s</i>" % _("Canceling..."))
    self.button_cancel_test.setEnabled(False)
    self.dialog.setValue(100)

  def select_mirror(self, mirror):
    """Select and expand the path to a matching mirror in the list"""
    found = self.treeview.findItems(mirror, Qt.MatchExactly|Qt.MatchRecursive)
    if found:
      found[0].setSelected(True)
      self.treeview.setCurrentItem(found[0])
      self.treeview.scrollToItem(found[0], QAbstractItemView.PositionAtCenter)
      self.on_treeview_mirrors_cursor_changed(found[0], 0)
      self.button_choose.setFocus()
      return True
